import base64

import pandas as pd

BASE64_PREFIX = 'base64:'


class Base64Mixin:
    def convert_data_format(self, data):
        """Convert byte data to base64 format for storage in the database."""
        if isinstance(data, bytes):
            return BASE64_PREFIX + base64.b64encode(data).decode('utf-8')
        else:
            return data

    def recover_data_format(self, data):
        """Recover byte data from base64 format stored in the database."""
        if isinstance(data, str) and data.startswith(BASE64_PREFIX):
            return base64.b64decode(data[len(BASE64_PREFIX) :])
        else:
            return data

    def process_schema_types(self, schema_mapping):
        """Convert bytes to string in the schema."""
        for key, value in schema_mapping.items():
            if value == 'Bytes':
                schema_mapping[key] = 'String'
        return schema_mapping


class DBHelper:
    match_dialect = 'base'

    def __init__(self, dialect):
        self.dialect = dialect

    def process_before_insert(self, table_name, datas):
        return table_name, pd.DataFrame(datas)

    def process_schema_types(self, schema_mapping):
        return schema_mapping

    def convert_data_format(self, data):
        return data

    def recover_data_format(self, data):
        return data


class ClickHouseHelper(Base64Mixin, DBHelper):
    match_dialect = 'clickhouse'

    def process_before_insert(self, table_name, datas):
        return f'`{table_name}`', pd.DataFrame(datas)


def get_db_helper(dialect) -> DBHelper:
    """Get the insert processor for the given dialect."""
    for helper in DBHelper.__subclasses__():
        if helper.match_dialect == dialect:
            return helper(dialect)

    return DBHelper(dialect)



import typing as t

import ray

from aladindb import logging
from aladindb.backends.base.compute import ComputeBackend


class RayComputeBackend(ComputeBackend):
    """
    A client for interacting with a ray cluster. Initialize the ray client.

    :param address: The address of the ray cluster.
    :param local: Set to True to create a local Dask cluster. (optional)
    :param **kwargs: Additional keyword arguments to be passed to the ray client.
    """

    def __init__(
        self,
        address: t.Optional[str] = None,
        local: bool = False,
        **kwargs,
    ):
        self._futures_collection: t.Dict[str, ray.ObjectRef] = {}
        if local:
            ray.init()
        else:
            ray.init(address=address, **kwargs)

    @property
    def type(self) -> str:
        return "distributed"

    @property
    def name(self) -> str:
        return "ray"

    def submit(self, function: t.Callable, *args, **kwargs) -> ray.ObjectRef:
        """
        Submits a function to the ray server for execution.

        :param function: The function to be executed.
        """

        def _dependable_remote_job(function, *args, **kwargs):
            if (
                function.__name__ in ['method_job', 'callable_job']
                and 'dependencies' in kwargs
            ):
                dependencies = kwargs['dependencies']
                if dependencies:
                    ray.wait(dependencies)
            return function(*args, **kwargs)

        remote_function = ray.remote(_dependable_remote_job)
        future = remote_function.remote(function, *args, **kwargs)
        self._futures_collection[future.task_id().hex()] = future

        logging.success(f"Job submitted.  function:{function} future:{future}")
        return future

    @property
    def tasks(self) -> t.Dict[str, ray.ObjectRef]:
        """
        List all pending tasks
        """
        return self._futures_collection

    def wait(self, identifier: str) -> None:
        """
        Waits for task corresponding to identifier to complete.
        :param identifier: Future task id to wait
        """
        ray.wait([self._futures_collection[identifier]])

    def wait_all(self) -> None:
        """
        Waits for all tasks to complete.
        """
        ray.wait(
            list(self._futures_collection.values()),
            num_returns=len(self._futures_collection),
        )

    def result(self, identifier: str) -> t.Any:
        """
        Retrieves the result of a previously submitted task.
        Note: This will block until the future is completed.

        :param identifier: The identifier of the submitted task.
        """
        future = self._futures_collection[identifier]
        return ray.get(future)

    def disconnect(self) -> None:
        """
        Disconnect the ray client.
        """
        ray.shutdown()

    def shutdown(self) -> None:
        """
        Shuts down the ray cluster.
        """
        ray.shutdown()



import dataclasses as dc
import typing as t

from aladindb.base.artifact import Artifact
from aladindb.components.component import Component
from aladindb.misc.annotations import public_api


@public_api(stability='beta')
@dc.dataclass(kw_only=True)
class Metric(Component):
    """
    Metric base object with which to evaluate performance on a data-set.
    These objects are ``callable`` and are applied row-wise to the data, and averaged.
    {component_parameters}
    :param object: callable or ``Artifact`` to be applied to the data
    """

    __doc__ = __doc__.format(component_parameters=Component.__doc__)

    artifacts: t.ClassVar[t.List[str]] = ['object']
    type_id: t.ClassVar[str] = 'metric'

    object: t.Union[Artifact, t.Callable]

    def __post_init__(self) -> None:
        super().__post_init__()
        if self.object and not isinstance(self.object, Artifact):
            self.object = Artifact(artifact=self.object)

    def __call__(self, x: t.Sequence[int], y: t.Sequence[int]) -> bool:
        assert isinstance(self.object, Artifact)
        return self.object.artifact(x, y)



import hashlib
import os
import re
import signal
import sys
import tempfile
import typing as t
import warnings
import zipfile
from contextlib import contextmanager
from io import BytesIO
from multiprocessing.pool import ThreadPool

import boto3
import requests
from tqdm import tqdm

from aladindb import CFG, logging
from aladindb.backends.base.query import Insert, Select
from aladindb.base.document import Document
from aladindb.base.serializable import Serializable


class TimeoutException(Exception):
    ...


def timeout_handler(signum, frame):
    raise TimeoutException()


@contextmanager
def timeout(seconds):
    old_handler = signal.signal(signal.SIGALRM, timeout_handler)
    signal.alarm(seconds)
    try:
        yield
    finally:
        signal.alarm(0)
        signal.signal(signal.SIGALRM, old_handler)


class Fetcher:
    """
    Fetches data from a URI

    :param headers: headers to be used for download
    :param n_workers: number of download workers
    """

    def __init__(self, headers: t.Optional[t.Dict] = None, n_workers: int = 0):
        session = boto3.Session()
        self.headers = headers
        self.s3_client = session.client("s3")
        self.request_session = requests.Session()
        self.request_adapter = requests.adapters.HTTPAdapter(
            max_retries=3,
            pool_connections=n_workers if n_workers else 1,
            pool_maxsize=n_workers * 10,
        )
        self.request_session.mount("http://", self.request_adapter)
        self.request_session.mount("https://", self.request_adapter)

    def _download_s3_object(self, uri):
        f = BytesIO()
        path = uri.split('s3://')[-1]
        bucket_name = path.split('/')[0]
        file = '/'.join(path.split('/')[1:])
        self.s3_client.download_fileobj(bucket_name, file, f)
        return f.getvalue()

    def _download_file(self, path):
        path = re.split('^file://', path)[-1]
        with open(path, 'rb') as f:
            return f.read()

    def _download_from_uri(self, uri):
        return self.request_session.get(uri, headers=self.headers).content

    def __call__(self, uri: str):
        """
        Download data from a URI

        :param uri: uri to download from
        """
        if uri.startswith('file://'):
            return self._download_file(uri)
        elif uri.startswith('s3://'):
            return self._download_s3_object(uri)
        elif uri.startswith('http://') or uri.startswith('https://'):
            return self._download_from_uri(uri)
        else:
            raise NotImplementedError(f'unknown type of URI "{uri}"')


class BaseDownloader:
    """
    Base class for downloading files

    :param uris: list of uris/ file names to fetch
    :param n_workers: number of multiprocessing workers
    :param timeout: set seconds until request times out
    :param headers: dictionary of request headers passed to``requests`` package
    :param raises: raises error ``True``/``False``
    """

    def __init__(
        self,
        uris: t.List[str],
        n_workers: int = 0,
        timeout: t.Optional[int] = None,
        headers: t.Optional[t.Dict] = None,
        raises: bool = True,
    ):
        self.timeout = timeout
        self.n_workers = n_workers
        self.uris = uris
        self.headers = headers or {}
        self.raises = raises

    def go(self):
        """
        Download all files
        Uses a :py:class:`multiprocessing.pool.ThreadPool` to parallelize
                          connections.
        """
        logging.info(f'number of workers {self.n_workers}')
        prog = tqdm(total=len(self.uris))
        prog.prefix = 'downloading from uris'
        self.failed = 0
        prog.prefx = "failed: 0"

        def f(i):
            prog.update()
            try:
                if self.timeout is not None:
                    with timeout(self.timeout):
                        self._download(i)
                else:
                    self._download(i)
            except TimeoutException:
                logging.warning(f'timed out {i}')
            except Exception as e:
                if self.raises:
                    raise e
                warnings.warn(str(e))
                self.failed += 1
                prog.prefix = f"failed: {self.failed} [{e}]"

        if self.n_workers == 0:
            self._sequential_go(f)
            return

        self._parallel_go(f)

    def _check_exists_if_hybrid(self, uri):
        if uri.startswith('file://'):
            file = f'{CFG.downloads_folder}/{uri.split("file://")[-1]}'
        else:
            file = f'{CFG.downloads_folder}/{hashlib.sha1(uri.encode()).hexdigest()}'
        if os.path.exists(file):
            return True
        return False

    def _parallel_go(self, f):
        pool = ThreadPool(self.n_workers)
        try:
            pool.map(f, range(len(self.uris)))
        except KeyboardInterrupt:
            logging.warning("--keyboard interrupt--")
            pool.terminate()
            pool.join()
            sys.exit(1)  # Kill this subprocess so it doesn't hang

        pool.close()
        pool.join()

    def _sequential_go(self, f):
        for i in range(len(self.uris)):
            f(i)


class SaveFile:
    """
    Save file to disk.

    :param root: root directory to save files to.
    """

    def __init__(self, root: str):
        self.root = root

    def __call__(self, bytes_: bytearray, uri: str, **kwargs) -> None:
        path = f'{self.root}/{hashlib.sha1(uri.encode()).hexdigest()}'
        with open(path, 'wb') as f:
            f.write(bytes_)


class Downloader(BaseDownloader):
    """
    Download files from a list of URIs.

    :param uris: list of uris/ file names to fetch
    :param update_one: function to call to insert data into table
    :param ids: list of ids of rows/ documents to update
    :param keys: list of keys in rows/ documents to insert to
    :param n_workers: number of multiprocessing workers
    :param headers: dictionary of request headers passed to``requests`` package
    :param skip_existing: if ``True`` then don't bother getting already present data
    :param timeout: set seconds until request times out
    :param raises: raises error ``True``/``False``
    """

    results: t.Dict[int, str]

    def __init__(
        self,
        uris,
        update_one: t.Optional[t.Callable] = None,
        ids: t.Optional[t.Union[t.List[str], t.List[int]]] = None,
        keys: t.Optional[t.List[str]] = None,
        n_workers: int = 20,
        headers: t.Optional[t.Dict] = None,
        skip_existing: bool = True,
        timeout: t.Optional[int] = None,
        raises: bool = True,
    ):
        super().__init__(
            uris, n_workers=n_workers, timeout=timeout, headers=headers, raises=raises
        )

        if ids is not None:
            if len(ids) != len(uris):
                raise ValueError(f'len(ids={ids}) != len(uris={uris})')

        self.ids = ids
        self.keys = keys
        self.failed = 0
        self.skip_existing = skip_existing
        self.update_one = update_one
        self.fetcher = Fetcher(headers=headers, n_workers=n_workers)

    def _download(self, i):
        if CFG.hybrid_storage:
            if self._check_exists_if_hybrid(self.uris[i]):
                return
        content = self.fetcher(self.uris[i])
        self.update_one(
            id=self.ids[i],
            key=self.keys[i],
            bytes_=content,
            uri=self.uris[i],
        )


def gather_uris(
    documents: t.Sequence[t.Dict], gather_ids: bool = True
) -> t.Tuple[t.List[str], t.List[str], t.Union[t.List[int], t.List[str]]]:
    """
    Get the uris out of all documents as denoted by ``{"_content": ...}``

    :param documents: list of dictionaries
    :param gather_ids: if ``True`` then gather ids of documents
    """
    uris = []
    mongo_keys = []
    ids = []
    for i, r in enumerate(documents):
        sub_uris, sub_mongo_keys = _gather_uris_for_document(r)
        if gather_ids:
            ids.extend([r['_id'] for _ in sub_uris])
        else:
            ids.append(i)
        uris.extend(sub_uris)
        mongo_keys.extend(sub_mongo_keys)
    return uris, mongo_keys, ids


def _gather_uris_for_document(r: t.Dict):
    '''
    >>> _gather_uris_for_document({'a': {'_content': {'uri': 'test'}}})
    (['test'], ['a'])
    >>> d = {'b': {'a': {'_content': {'uri': 'test'}}}}
    >>> _gather_uris_for_document(d)
    (['test'], ['b.a'])
    >>> d = {'b': {'a': {'_content': {'uri': 'test', 'bytes': b'abc'}}}}
    >>> _gather_uris_for_document(d)
    ([], [])
    '''
    uris = []
    keys = []
    for k in r:
        if isinstance(r[k], dict) and '_content' in r[k]:
            if 'uri' in r[k]['_content'] and 'bytes' not in r[k]['_content']:
                keys.append(k)
                uris.append(r[k]['_content']['uri'])
        elif isinstance(r[k], dict) and '_content' not in r[k]:
            sub_uris, sub_keys = _gather_uris_for_document(r[k])
            uris.extend(sub_uris)
            keys.extend([f'{k}.{key}' for key in sub_keys])
    return uris, keys


def download_content(
    db,
    query: t.Union[Select, Insert, t.Dict],
    ids: t.Optional[t.Sequence[str]] = None,
    documents: t.Optional[t.List[Document]] = None,
    timeout: t.Optional[int] = None,
    raises: bool = True,
    n_download_workers: t.Optional[int] = None,
    headers: t.Optional[t.Dict] = None,
    download_update: t.Optional[t.Callable] = None,
    **kwargs,
) -> t.Optional[t.Sequence[Document]]:
    """
    Download content contained in uploaded data. Items to be downloaded are identifier
    via the subdocuments in the form exemplified below. By default items are downloaded
    to the database, unless a ``download_update`` function is provided.

    :param db: database instance
    :param query: query to be executed
    :param ids: ids to be downloaded
    :param documents: documents to be downloaded
    :param timeout: timeout for download
    :param raises: whether to raise errors
    :param n_download_workers: number of download workers
    :param headers: headers to be used for download
    :param download_update: function to be used for updating the database
    :param **kwargs: additional keyword arguments

    >>> d = {"_content": {"uri": "<uri>", "encoder": "<encoder-identifier>"}}
    >>> def update(key, id, bytes):
    >>> ... with open(f'/tmp/{key}+{id}', 'wb') as f:
    >>> ...     f.write(bytes)
    >>> download_content(None, None, ids=["0"], documents=[d]))
    ...
    """
    logging.debug(query.__str__())
    logging.debug(ids.__str__())
    update_db = False
    if isinstance(query, dict):
        query = Serializable.deserialize(query)

    if documents is not None:
        pass
    elif isinstance(query, Select):
        update_db = True
        if ids is None:
            documents = list(db.execute(query).raw_cursor)
        else:
            select = query.select_using_ids(ids)
            documents = list(db.execute(select))
    else:
        assert isinstance(query, Insert)
        documents = t.cast(t.List[Document], query.documents)

    uris, keys, place_ids = gather_uris([d.encode() for d in documents])
    if uris:
        logging.info(f'found {len(uris)} uris')
    if not uris:
        return None

    if n_download_workers is None:
        try:
            n_download_workers = db.metadata.get_metadata(key='n_download_workers')
        except TypeError:
            n_download_workers = 0

    if headers is None:
        try:
            headers = db.metadata.get_metadata(key='headers')
        except TypeError:
            pass

    if timeout is None:
        try:
            timeout = db.metadata.get_metadata(key='download_timeout')
        except TypeError:
            pass

    if CFG.hybrid_storage:
        assert isinstance(CFG.downloads_folder, str)
        _download_update = SaveFile(CFG.downloads_folder)
    else:

        def _download_update(key, id, bytes_, **kwargs):  # type: ignore[misc]
            return query.download_update(db=db, key=key, id=id, bytes=bytes_)

    assert place_ids is not None
    downloader = Downloader(
        uris=uris,
        ids=place_ids,
        keys=keys,
        update_one=download_update or _download_update,
        n_workers=n_download_workers,
        timeout=timeout,
        headers=headers,
        raises=raises,
    )
    downloader.go()
    if update_db:
        return None

    for id_, key in zip(place_ids, keys):
        documents[id_] = db.db.set_content_bytes(  # type: ignore[call-overload]
            documents[id_],  # type: ignore[call-overload]
            key,
            downloader.results[id_],  # type: ignore[index]
        )
    return documents


def remote_archive(uri, fetcher):
    with tempfile.TemporaryDirectory() as tmp_dir:
        tmp_path = tmp_dir / 'archive.zip'
        logging.info(f'downloading {uri} to {tmp_path}')
        fetcher(uri, path=tmp_path)

        zipfile.ZipFile(tmp_path).extractall(tmp_dir / 'extracted')
        assert 'artifact' in os.listdir(tmp_dir / 'extracted')

        with open(tmp_dir / 'extracted' / 'artifact', 'rb') as f:
            return f.read()



import dataclasses as dc
import json
import os
import typing as t

from bson import objectid

from aladindb.base.serializable import Serializable
from aladindb.cdc.cdc import DBEvent, Packet

TokenType = t.Dict[str, str]


class CachedTokens:
    token_path = os.path.join('.aladindb', '.cdc.tokens')
    separate = '\n'

    def __init__(self):
        # BROKEN: self._current_tokens is never read from
        self._current_tokens = []

    def append(self, token: TokenType) -> None:
        with open(CachedTokens.token_path, 'a') as fp:
            stoken = json.dumps(token)
            stoken = stoken + self.separate
            fp.write(stoken)

    def load(self) -> t.Sequence[TokenType]:
        with open(CachedTokens.token_path) as fp:
            tokens = fp.read().split(self.separate)[:-1]
            self._current_tokens = [TokenType(json.loads(t)) for t in tokens]
        return self._current_tokens


class ObjectId(objectid.ObjectId):
    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):
        if not isinstance(v, objectid.ObjectId):
            raise TypeError('Id is required.')
        return str(v)


@dc.dataclass
class MongoDBPacket(Packet):
    """
    A base packet to represent message in task queue.
    """

    ids: t.List[t.Union[ObjectId, str]]
    query: t.Optional[Serializable] = None
    event_type: DBEvent = DBEvent.insert



"""
The component module provides the base class for all components in aladindb.
"""


from __future__ import annotations

import dataclasses as dc
import typing as t

from aladindb.backends.base.artifact import ArtifactStore
from aladindb.base.serializable import Serializable
from aladindb.jobs.job import ComponentJob, Job

if t.TYPE_CHECKING:
    from aladindb.base.datalayer import Datalayer
    from aladindb.components.dataset import Dataset


@dc.dataclass
class Component(Serializable):
    """
    :param identifier: A unique identifier for the component"""

    type_id: t.ClassVar[str]
    identifier: str

    @property
    def db(self):
        return self._db

    @db.setter
    def db(self, value: Datalayer):
        self._db = value

    def __post_init__(self) -> None:
        # set version in `__post_init__` so that is
        # cannot be set in `__init__` and is always set
        self.version: t.Optional[int] = None
        self._db = None  # type: ignore[assignment]
        if not self.identifier:
            raise ValueError('identifier cannot be empty or None')

    def pre_create(self, db: Datalayer) -> None:
        """Called the first time this component is created

        :param db: the db that creates the component
        """
        assert db

    def post_create(self, db: Datalayer) -> None:
        """Called after the first time this component is created.
        Generally used if ``self.version`` is important in this logic.

        :param db: the db that creates the component
        """
        assert db

    def on_load(self, db: Datalayer) -> None:
        """Called when this component is loaded from the data store

        :param db: the db that loaded the component
        """
        assert db

    @property
    def child_components(self) -> t.Sequence[t.Any]:
        return []

    @property
    def serialized(self):
        serialized = self.serialize()
        artifacts = tuple(set(ArtifactStore.get_artifacts(serialized)))
        return serialized, artifacts

    @property
    def unique_id(self) -> str:
        if getattr(self, 'version', None) is None:
            raise Exception('Version not yet set for component uniqueness')
        return f'{self.type_id}/' f'{self.identifier}/' f'{self.version}'

    def create_validation_job(
        self,
        validation_set: t.Union[str, Dataset],
        metrics: t.Sequence[str],
    ) -> ComponentJob:
        assert self.identifier is not None
        return ComponentJob(
            component_identifier=self.identifier,
            method_name='predict',
            type_id='model',
            kwargs={
                'validation_set': validation_set,
                'metrics': metrics,
            },
        )

    def schedule_jobs(
        self,
        db: Datalayer,
        dependencies: t.Sequence[Job] = (),
        verbose: bool = False,
    ) -> t.Sequence[t.Any]:
        """Run the job for this listener

        :param database: The db to process
        :param dependencies: A sequence of dependencies,
        :param verbose: If true, print more information
        """
        return []

    @classmethod
    def make_unique_id(cls, type_id: str, identifier: str, version: int) -> str:
        return f'{type_id}/{identifier}/{version}'



import datetime
import typing as t
import uuid
from abc import abstractmethod

import aladindb as s
from aladindb.jobs.tasks import callable_job, method_job


def job(f):
    def wrapper(
        *args,
        db: t.Any = None,
        dependencies: t.Sequence[Job] = (),
        **kwargs,
    ):
        j = FunctionJob(callable=f, args=args, kwargs=kwargs)
        return j(db=db, dependencies=dependencies)

    return wrapper


class Job:
    """
    Base class for jobs. Jobs are used to run functions or methods on.

    :param args: positional arguments to be passed to the function or method
    :param kwargs: keyword arguments to be passed to the function or method
    :param identifier: unique identifier
    :param callable: function or method to be called
    :param db: DB instance to be used
    :param future: future object returned by dask
    """

    callable: t.Optional[t.Callable]

    def __init__(
        self,
        args: t.Optional[t.Sequence] = None,
        kwargs: t.Optional[t.Dict] = None,
    ):
        self.args = args or ()
        self.kwargs = kwargs or {}
        self.identifier = str(uuid.uuid4())
        self.time = datetime.datetime.now()
        self.callable = None
        self.db = None
        self.future = None

    def watch(self):
        """
        Watch the stdout of the job.
        """
        return self.db.metadata.watch_job(identifier=self.identifier)

    @abstractmethod
    def submit(self, compute, dependencies=()):
        """
        Submit job for execution

        :param compute: compute engine
        :param dependencies: list of dependencies
        """
        raise NotImplementedError

    def dict(self):
        return {
            'identifier': self.identifier,
            'time': self.time,
            'status': 'pending',
            'args': self.args,
            'kwargs': self.kwargs,
            'stdout': [],
            'stderr': [],
        }

    def __call__(self, db: t.Any = None, dependencies=()):
        """
        Run the job.

        :param db: DB instance to be used
        :param dependencies: list of dependencies
        """
        raise NotImplementedError


class FunctionJob(Job):
    """
    Job for running a function.
    on a dask cluster.

    :param callable: function to be called
    :param args: positional arguments to be passed to the function
    :param kwargs: keyword arguments to be passed to the function
    """

    def __init__(
        self,
        callable: t.Callable,
        args: t.Optional[t.Sequence] = None,
        kwargs: t.Optional[t.Dict] = None,
    ):
        super().__init__(args=args, kwargs=kwargs)
        self.callable = callable

    def dict(self):
        d = super().dict()
        d['cls'] = 'FunctionJob'
        return d

    def submit(self, dependencies=()):
        """
        Submit job for execution

        :param dependencies: list of dependencies
        """
        self.future = self.db.compute.submit(
            callable_job,
            cfg=s.CFG.dict(),
            function_to_call=self.callable,
            job_id=self.identifier,
            args=self.args,
            kwargs=self.kwargs,
            dependencies=dependencies,
            db=self.db if self.db.compute.type == 'local' else None,
            local=self.db.compute.type == 'local',
        )

        return

    def __call__(self, db: t.Any = None, dependencies=()):
        if db is None:
            from aladindb.base.build import build_datalayer

            db = build_datalayer()

        self.db = db
        db.metadata.create_job(self.dict())

        self.submit(dependencies=dependencies)
        return self


class ComponentJob(Job):
    """
    Job for running a class method of a component.

    :param component_identifier: unique identifier of the component
    :param type_id: type of the component
    :param method_name: name of the method to be called
    :param args: positional arguments to be passed to the method
    :param kwargs: keyword arguments to be passed to the method
    """

    def __init__(
        self,
        component_identifier: str,
        type_id: str,
        method_name: str,
        args: t.Optional[t.Sequence] = None,
        kwargs: t.Optional[t.Dict] = None,
    ):
        super().__init__(args=args, kwargs=kwargs)

        self.component_identifier = component_identifier
        self.method_name = method_name
        self.type_id = type_id
        self._component = None

    @property
    def component(self):
        return self._component

    @component.setter
    def component(self, value):
        self._component = value
        self.callable = getattr(self._component, self.method_name)

    def submit(self, dependencies=()):
        """
        Submit job for execution
        :param dependencies: list of dependencies
        """
        self.future = self.db.compute.submit(
            method_job,
            cfg=s.CFG.dict(),
            type_id=self.type_id,
            identifier=self.component_identifier,
            method_name=self.method_name,
            job_id=self.identifier,
            args=self.args,
            kwargs=self.kwargs,
            dependencies=dependencies,
            db=self.db if self.db.compute.type == 'local' else None,
            local=self.db.compute.type == 'local',
        )
        return

    def __call__(self, db: t.Any = None, dependencies=()):
        if db is None:
            from aladindb.base.build import build_datalayer

            db = build_datalayer()

        self.db = db

        db.metadata.create_job(self.dict())
        if self.component is None:
            self.component = db.load(self.type_id, self.component_identifier)
        self.submit(dependencies=dependencies)
        return self

    def dict(self):
        d = super().dict()
        d.update(
            {
                'method_name': self.method_name,
                'component_identifier': self.component_identifier,
                'type_id': self.type_id,
                'cls': 'ComponentJob',
            }
        )
        return d



from typer import Typer

from .. import ICON

__all__ = 'app', 'command'

CLI_NAME = 'aladindb'

app = Typer(
    add_completion=False,
    context_settings={'help_option_names': ['--help', '-h']},
    help=f"""\
{ICON} {CLI_NAME} {ICON}

Usage: {CLI_NAME} [GLOBAL-FLAGS] [COMMAND] [COMMAND-FLAGS] [COMMAND-ARGS]

  Examples:

    $ {CLI_NAME} serve

    $ {CLI_NAME} local-cluster users documents products

    $ {CLI_NAME} config
""",
)

command = app.command



import subprocess

from . import command


@command(help='Start local dask scheduler')
def dask_scheduler():
    subprocess.Popen(["dask", "scheduler"])


@command(help='Start local dask worker')
def dask_worker():
    subprocess.Popen(["dask", "worker", "tcp://localhost:8786"])


@command(help='Start local cluster: server, dask and change data capture')
def local_cluster():
    from aladindb.base.build import build_datalayer
    from aladindb.server.cluster import cluster

    db = build_datalayer()
    cluster(db)


@command(help='Start vector search server')
def vector_search():
    from aladindb.vector_search.server.app import app

    app.start()


@command(help='Start standalone change data capture')
def cdc():
    from aladindb.cdc.app import app

    app.start()



import re
import sys
import typing as t

import ibis
import mongomock
import pandas
import pymongo

import aladindb as s
from aladindb import logging
from aladindb.backends.base.backends import data_backends, metadata_stores
from aladindb.backends.local.artifacts import FileSystemArtifactStore
from aladindb.backends.local.compute import LocalComputeBackend
from aladindb.backends.mongodb.artifacts import MongoArtifactStore
from aladindb.backends.ray.compute import RayComputeBackend
from aladindb.base.datalayer import Datalayer


def build_metadata(metadata_store=None):
    if metadata_store is None:
        metadata_store = s.CFG.metadata_store
    return build(metadata_store, metadata_stores, type='metadata')


def build_databackend(databackend: t.Optional[str] = None):
    if databackend is None:
        databackend = s.CFG.data_backend
    return build(databackend, data_backends)


def build_artifact_store(artifact_store: str):
    if artifact_store.startswith('mongodb://'):
        import pymongo

        conn: pymongo.MongoClient = pymongo.MongoClient(
            '/'.join(artifact_store.split('/')[:-1])
        )
        name = artifact_store.split('/')[-1]
        return MongoArtifactStore(conn, name)
    elif artifact_store.startswith('filesystem://'):
        directory = artifact_store.split('://')[1]
        return FileSystemArtifactStore(directory)
    else:
        raise ValueError(f'Unknown artifact store: {artifact_store}')


# Helper function to build a data backend based on the URI.
def build(uri, mapping, type: str = 'data_backend'):
    logging.debug(f"Parsing data connection URI:{uri}")

    if re.match('^mongodb:\/\/', uri) is not None:
        name = uri.split('/')[-1]
        conn: pymongo.MongoClient = pymongo.MongoClient(
            uri,
            serverSelectionTimeoutMS=5000,
        )
        return mapping['mongodb'](conn, name)

    elif re.match('^mongodb\+srv:\/\/', uri):
        name = uri.split('/')[-1]
        conn = pymongo.MongoClient(
            '/'.join(uri.split('/')[:-1]),
            serverSelectionTimeoutMS=5000,
        )
        return mapping['mongodb'](conn, name)

    elif uri.startswith('mongomock://'):
        name = uri.split('/')[-1]
        conn = mongomock.MongoClient()
        return mapping['mongodb'](conn, name)

    elif uri.endswith('.csv'):
        if type == 'metadata':
            raise ValueError('Cannot build metadata from a CSV file.')

        import glob

        csv_files = glob.glob(uri)
        tables = {}
        for csv_file in csv_files:
            pattern = re.match('^.*/(.*)\.csv$', csv_file)
            assert pattern is not None
            tables[pattern.groups()[0]] = pandas.read_csv(csv_file)
        ibis_conn = ibis.pandas.connect(tables)
        return mapping['ibis'](ibis_conn, uri.split('/')[0])
    else:
        name = uri.split('//')[0]
        if type == 'data_backend':
            ibis_conn = ibis.connect(uri)
            return mapping['ibis'](ibis_conn, name)
        else:
            assert type == 'metadata'
            from sqlalchemy import create_engine

            sql_conn = create_engine(uri)
            return mapping['sqlalchemy'](sql_conn, name)


def build_compute(compute):
    if compute == 'local' or compute is None:
        return LocalComputeBackend()

    if compute == 'dask+thread':
        from aladindb.backends.dask.compute import DaskComputeBackend

        return DaskComputeBackend('', local=True)

    if compute.split('://')[0] == 'dask+tcp':
        from aladindb.backends.dask.compute import DaskComputeBackend

        uri = compute.split('+')[-1]
        return DaskComputeBackend(uri)

    if compute.split('://')[0] == 'ray':
        return RayComputeBackend(compute)

    return LocalComputeBackend()


def build_datalayer(cfg=None, databackend=None, **kwargs) -> Datalayer:
    """
    Build a Datalayer object as per ``db = aladin(db)`` from configuration.

    :param cfg: Configuration to use. If None, use ``aladindb.CFG``.
    :param databackend: Databacked to use.
                        If None, use ``aladindb.CFG.data_backend``.
    """

    # Configuration
    # ------------------------------
    # Use the provided configuration or fall back to the default configuration.
    cfg = cfg or s.CFG

    # Update configuration with keyword arguments.
    for k, v in kwargs.items():
        cfg.force_set(k, v)

    # Connect to data backend.
    # ------------------------------
    try:
        if not databackend:
            databackend = build(cfg.data_backend, data_backends)
        logging.info("Data Client is ready.", databackend.conn)
    except Exception as e:
        # Exit quickly if a connection fails.
        logging.error("Error initializing to DataBackend Client:", str(e))
        sys.exit(1)

    # Connect to metadata store.
    # ------------------------------
    # 1. try to connect to the metadata store specified in the configuration.
    # 2. if that fails, try to connect to the data backend engine.
    # 3. if that fails, try to connect to the data backend uri.
    if cfg.metadata_store is not None:
        # try to connect to the metadata store specified in the configuration.
        logging.info("Connecting to Metadata Client:", cfg.metadata_store)
        metadata = build(cfg.metadata_store, metadata_stores, type='metadata')
    else:
        try:
            # try to connect to the data backend engine.
            logging.info(
                "Connecting to Metadata Client with engine: ", databackend.conn
            )
            metadata = databackend.build_metadata()
        except Exception as e:
            logging.warn("Error building metadata from DataBackend:", str(e))
            metadata = None

    if metadata is None:
        try:
            # try to connect to the data backend uri.
            logging.info("Connecting to Metadata Client with URI: ", cfg.data_backend)
            metadata = build(cfg.data_backend, metadata_stores, type='metadata')
        except Exception as e:
            # Exit quickly if a connection fails.
            logging.error("Error initializing to Metadata Client:", str(e))
            sys.exit(1)

    # Build DataLayer
    # ------------------------------
    db = Datalayer(
        databackend=databackend,
        metadata=metadata,
        artifact_store=(
            build_artifact_store(cfg.artifact_store)
            if cfg.artifact_store is not None
            else databackend.build_artifact_store()
        ),
        compute=build_compute(cfg.cluster.compute),
    )

    return db



import copy
import dataclasses as dc
import types
import typing as t

_ATOMIC_TYPES = frozenset(
    {
        # Common JSON Serializable types
        type(None),
        bool,
        int,
        float,
        str,
        # Other common types
        complex,
        bytes,
        # Other types that are also unaffected by deepcopy
        type(...),
        type(NotImplemented),
        types.CodeType,
        types.BuiltinFunctionType,
        types.FunctionType,
        type,
        range,
        property,
    }
)


def asdict(obj, *, copy_method=copy.copy) -> t.Dict[str, t.Any]:
    """
    Custom ``asdict`` function which exports a dataclass object into a dict,
    with a option to choose for nested non atomic objects copy strategy.
    """
    if not dc.is_dataclass(obj):
        raise TypeError("asdict() should be called on dataclass instances")
    return _asdict_inner(obj, dict, copy_method)


def _asdict_inner(obj, dict_factory, copy_method) -> t.Any:
    if type(obj) in _ATOMIC_TYPES:
        return obj
    elif dc.is_dataclass(obj):
        # fast path for the common case
        return {
            f.name: _asdict_inner(getattr(obj, f.name), dict, copy_method)
            for f in dc.fields(obj)
        }
    elif isinstance(obj, tuple) and hasattr(obj, '_fields'):
        # obj is a namedtuple.  Recurse into it, but the returned
        # object is another namedtuple of the same type.  This is
        # similar to how other list- or tuple-derived classes are
        # treated (see below), but we just need to create them
        # differently because a namedtuple's __init__ needs to be
        # called differently (see bpo-34363).

        # I'm not using namedtuple's _asdict()
        # method, because:
        # - it does not recurse in to the namedtuple fields and
        #   convert them to dicts (using dict_factory).
        # - I don't actually want to return a dict here.  The main
        #   use case here is json.dumps, and it handles converting
        #   namedtuples to lists.  Admittedly we're losing some
        #   information here when we produce a json list instead of a
        #   dict.  Note that if we returned dicts here instead of
        #   namedtuples, we could no longer call asdict() on a data
        #   structure where a namedtuple was used as a dict key.

        return type(obj)(*[_asdict_inner(v, dict_factory, copy_method) for v in obj])
    elif isinstance(obj, (list, tuple)):
        # Assume we can create an object of this type by passing in a
        # generator (which is not true for namedtuples, handled
        # above).
        return type(obj)(_asdict_inner(v, dict_factory, copy_method) for v in obj)
    elif isinstance(obj, dict):
        if hasattr(type(obj), 'default_factory'):
            # obj is a defaultdict, which has a different constructor from
            # dict as it requires the default_factory as its first arg.
            result = type(obj)(getattr(obj, 'default_factory'))
            for k, v in obj.items():
                result[_asdict_inner(k, dict_factory, copy_method)] = _asdict_inner(
                    v, dict_factory, copy_method
                )
            return result
        return type(obj)(
            (
                _asdict_inner(k, dict_factory, copy_method),
                _asdict_inner(v, dict_factory, copy_method),
            )
            for k, v in obj.items()
        )
    else:
        return copy_method(obj)



import json
from typing import Tuple

from sqlalchemy import (
    Boolean,
    DateTime,
    Integer,
    String,
    Text,
    TypeDecorator,
)

DEFAULT_LENGTH = 255


class JsonMixin:
    """Mixin for JSON type columns.
    Converts dict to JSON strings before saving to database
    and converts JSON strings to dict when loading from database.
    """

    def process_bind_param(self, value, dialect):
        if value is not None:
            value = json.dumps(value)
        return value

    def process_result_value(self, value, dialect):
        if value is not None:
            value = json.loads(value)
        return value


class JsonAsString(JsonMixin, TypeDecorator):
    impl = String(DEFAULT_LENGTH)


class JsonAsText(JsonMixin, TypeDecorator):
    impl = Text


class DefaultConfig:
    type_string = String(DEFAULT_LENGTH)
    type_json_as_string = JsonAsString
    type_json_as_text = JsonAsText
    type_integer = Integer
    type_datetime = DateTime
    type_boolean = Boolean

    query_id_table_args: Tuple = tuple()
    job_table_args: Tuple = tuple()
    parent_child_association_table_args: Tuple = tuple()
    component_table_args: Tuple = tuple()
    meta_table_args: Tuple = tuple()


def create_clickhouse_config():
    # lazy import
    try:
        from clickhouse_sqlalchemy import engines, types
    except ImportError:
        raise ImportError(
            'The clickhouse_sqlalchemy package is required to use the '
            'clickhouse dialect. Please install it with pip install '
            'clickhouse-sqlalchemy'
        )

    class ClickHouseConfig:
        class JsonAsString(JsonMixin, TypeDecorator):
            impl = types.String

        class JsonAsText(JsonMixin, TypeDecorator):
            impl = types.String

        type_string = types.String
        type_json_as_string = JsonAsString
        type_json_as_text = JsonAsText
        type_integer = types.Int32
        type_datetime = types.DateTime
        type_boolean = types.Boolean

        # clickhouse need engine args to create table
        query_id_table_args = (engines.MergeTree(order_by='query_id'),)
        job_table_args = (engines.MergeTree(order_by='identifier'),)
        parent_child_association_table_args = (engines.MergeTree(order_by='parent_id'),)
        component_table_args = (engines.MergeTree(order_by='id'),)
        meta_table_args = (engines.MergeTree(order_by='key'),)

    return ClickHouseConfig


def get_db_config(dialect):
    if dialect == 'clickhouse':
        return create_clickhouse_config()
    else:
        return DefaultConfig



import dataclasses as dc
import typing as t

import tenacity

import aladindb as s

ExceptionTypes = t.Union[t.Type[BaseException], t.Tuple[t.Type[BaseException], ...]]


@dc.dataclass
class Retry:
    """
    Retry a function until it succeeds.

    This is a thin wrapper around the tenacity retry library, using our configs.
    :param exception_types: The exception types to retry on.
    :param cfg: The retry config.
    """

    exception_types: ExceptionTypes
    cfg: t.Optional[s.config.Retry] = None

    def __call__(self, f: t.Callable) -> t.Any:
        cfg = self.cfg or s.CFG.retries
        retry = tenacity.retry_if_exception_type(self.exception_types)
        stop = tenacity.stop_after_attempt(cfg.stop_after_attempt)
        wait = tenacity.wait_exponential(
            max=cfg.wait_max,
            min=cfg.wait_min,
            multiplier=cfg.wait_multiplier,
        )
        retrier = tenacity.retry(retry=retry, stop=stop, wait=wait)
        return retrier(f)



import typing as t


class MongoStyleDict(t.Dict[str, t.Any]):
    """
    Dictionary object mirroring how fields can be referred to and set in MongoDB.

    >>> d = MongoStyleDict({'a': {'b': 1}})
    >>> d['a.b']
    1

    t.Set deep fields directly with string keys:
    >>> d['a.c'] = 2
    >>> d
    {'a': {'b': 1, 'c': 2}}

    Parent keys should exist in order to set subfields:
    >>> d['a.d.e'] = 3
    Traceback (most recent call last):
    ...
    KeyError: 'd'
    """

    def __getitem__(self, key: str) -> t.Any:
        if '.' not in key:
            return super().__getitem__(key)
        else:
            try:
                return super().__getitem__(key)
            except KeyError:
                parts = key.split('.')
                parent = parts[0]
                child = '.'.join(parts[1:])
                sub = MongoStyleDict(self.__getitem__(parent))
                return sub[child]

    def __setitem__(self, key: str, value: t.Any) -> None:
        if '.' not in key:
            super().__setitem__(key, value)
        else:
            parent = key.split('.')[0]
            child = '.'.join(key.split('.')[1:])
            try:
                parent_item = MongoStyleDict(self[parent])
            except KeyError:
                parent_item = MongoStyleDict({})
            parent_item[child] = value
            self[parent] = parent_item



import typing as t

import bson
from bson.objectid import ObjectId

from aladindb import CFG
from aladindb.base.config import BytesEncoding
from aladindb.components.encoder import Encodable, Encoder
from aladindb.components.schema import Schema
from aladindb.misc.files import get_file_from_uri
from aladindb.misc.special_dicts import MongoStyleDict

ContentType = t.Union[t.Dict, Encodable]
ItemType = t.Union[t.Dict[str, t.Any], Encodable, ObjectId]

_OUTPUTS_KEY: str = '_outputs'


class Document:
    """
    A wrapper around an instance of dict or a Encodable which may be used to dump
    that resource to a mix of JSONable and `bytes`

    :param content: The content to wrap
    """

    _DEFAULT_ID_KEY: str = '_id'

    content: ContentType

    def __init__(self, content: ContentType):
        self.content = content

    def dump_bson(self) -> bytes:
        """Dump this document into BSON and encode as bytes"""
        return bson.encode(self.encode())

    def encode(
        self,
        schema: t.Optional[Schema] = None,
        bytes_encoding: t.Optional[BytesEncoding] = None,
    ) -> t.Any:
        """Make a copy of the content with all the Encodables encoded"""
        bytes_encoding = bytes_encoding or CFG.bytes_encoding
        if schema is not None:
            return _encode_with_schema(self.content, schema, bytes_encoding)
        return _encode(self.content, bytes_encoding)

    @property
    def variables(self) -> t.List[str]:
        from aladindb.base.serializable import _find_variables

        return _find_variables(self.content)

    def set_variables(self, db, **kwargs) -> 'Document':
        from aladindb.base.serializable import _replace_variables

        content = _replace_variables(
            self.content, db, **kwargs
        )  # replace variables with values
        return Document(content)

    def outputs(self, key: str, model: str, version: t.Optional[int] = None) -> t.Any:
        """
        Get document ouputs on ``key`` from ``model``

        :param key: Document key to get outputs from.
        :param model: Model name to get outputs from.
        """
        r = MongoStyleDict(self.unpack())
        if version is not None:
            document = r[f'{_OUTPUTS_KEY}.{key}.{model}.{version}']
        else:
            tmp = r[f'{_OUTPUTS_KEY}.{key}.{model}']
            version = max(list(tmp.keys()))
            return tmp[version]
        return document

    @staticmethod
    def decode(
        r: t.Dict, encoders: t.Dict, bytes_encoding: t.Optional[BytesEncoding] = None
    ) -> t.Any:
        bytes_encoding = bytes_encoding or CFG.bytes_encoding

        if isinstance(r, Document):
            return Document(_decode(r, encoders, bytes_encoding))
        elif isinstance(r, dict):
            return _decode(r, encoders, bytes_encoding)
        raise NotImplementedError(f'type {type(r)} is not supported')

    def __repr__(self) -> str:
        return f'Document({repr(self.content)})'

    def __getitem__(self, item: str) -> ItemType:
        assert isinstance(self.content, dict)
        return self.content[item]

    def __setitem__(self, key: str, value: ItemType):
        assert isinstance(self.content, dict)
        self.content[key] = value

    def unpack(self) -> t.Any:
        """Returns the content, but with any encodables replacecs by their contents"""
        return _unpack(self.content)


def dump_bsons(documents: t.Sequence[Document]) -> bytes:
    """Dump a sequence of documents into BSON and encode as bytes

    :param documents: the sequence of documents to dump
    """
    return bytes(bson.encode({'docs': [d.encode() for d in documents]}))


def load_bson(content: t.ByteString, encoders: t.Dict[str, t.Any]) -> Document:
    """Load a Document from bson-encoded content

    :param content: the content to decode
    :param encoders: a dict of encoders
    """
    document: t.Dict = bson.decode(content)
    return Document(Document.decode(document, encoders=encoders))


def load_bsons(content: t.ByteString, encoders: t.Dict) -> t.List[Document]:
    """Load a list of Documents from bson-encoded content

    :param content: the content to decode
    :param encoders: a dict of encoders
    """
    d = t.cast(t.Dict, bson.decode(content))
    documents = d['docs']
    return [Document(Document.decode(r, encoders=encoders)) for r in documents]


def _decode(
    r: t.Dict, encoders: t.Dict, bytes_encoding: t.Optional[BytesEncoding] = None
) -> t.Any:
    bytes_encoding = bytes_encoding or CFG.bytes_encoding
    if isinstance(r, dict) and '_content' in r:
        encoder = encoders[r['_content']['encoder']]
        try:
            return encoder.decode(r['_content']['bytes'])
        except KeyError:
            if 'uri' in r['_content']:
                return Encodable(uri=r['_content']['uri'], encoder=encoder)
            return r
    elif isinstance(r, list):
        return [_decode(x, encoders) for x in r]
    elif isinstance(r, dict):
        for k in r:
            if k in encoders:
                r[k] = encoders[k].decode(r[k], bytes_encoding).x
            else:
                r[k] = _decode(r[k], encoders, bytes_encoding)
    return r


def _encode(r: t.Any, bytes_encoding: t.Optional[BytesEncoding] = None) -> t.Any:
    bytes_encoding = bytes_encoding or CFG.bytes_encoding

    if isinstance(r, dict):
        return {k: _encode(v, bytes_encoding) for k, v in r.items()}
    if isinstance(r, Encodable):
        return r.encode(bytes_encoding=bytes_encoding)
    return r


def _encode_with_schema(
    r: t.Any, schema: Schema, bytes_encoding: t.Optional[BytesEncoding] = None
) -> t.Any:
    bytes_encoding = bytes_encoding or CFG.bytes_encoding
    if isinstance(r, dict):
        out = {
            k: schema.fields[k].encode(v, wrap=False)  # type: ignore[call-arg]
            if isinstance(schema.fields[k], Encoder)
            else _encode_with_schema(v, schema, bytes_encoding)
            for k, v in r.items()
        }
        return out
    if isinstance(r, Encodable):
        return r.encode(bytes_encoding=bytes_encoding)
    return r


def _unpack(item: t.Any) -> t.Any:
    if isinstance(item, Encodable):
        if CFG.hybrid_storage and not item.encoder.load_hybrid and item.x is None:
            return get_file_from_uri(item.uri)
        return item.x
    elif isinstance(item, dict):
        return {k: _unpack(v) for k, v in item.items()}
    elif isinstance(item, list):
        return [_unpack(x) for x in item]
    else:
        return item



import typing as t

import numpy

from aladindb import logging
from aladindb.vector_search.base import BaseVectorSearcher, VectorItem, measures


class InMemoryVectorSearcher(BaseVectorSearcher):
    """
    Simple hash-set for looking up with vector similarity.

    :param identifier: Unique string identifier of index
    :param h: array/ tensor of vectors
    :param index: list of IDs
    :param measure: measure to assess similarity
    """

    name = 'vanilla'

    def __init__(
        self,
        identifier: str,
        dimensions: int,
        h: t.Optional[numpy.ndarray] = None,
        index: t.Optional[t.List[str]] = None,
        measure: t.Union[str, t.Callable] = 'cosine',
    ):
        self.identifier = identifier
        self.dimensions = dimensions

        if h is not None:
            assert index is not None
            self._setup(h, index)
        else:
            self.h_list = None
            self.h = None
            self.index = None
            self.lookup = None

        self.measure = measure
        if isinstance(measure, str):
            self.measure = measures[measure]

        self.identifier = identifier

    def __len__(self):
        return self.h.shape[0]

    def _setup(self, h, index):
        self.h_list = h if isinstance(h, list) else h.tolist()
        self.h = numpy.array(h) if not isinstance(h, numpy.ndarray) else h
        self.index = index
        self.lookup = dict(zip(index, range(len(index))))

    def find_nearest_from_id(self, _id, n=100):
        return self.find_nearest_from_array(self.h[self.lookup[_id]], n=n)

    def find_nearest_from_array(self, h, n=100, within_ids=None):
        h = self.to_numpy(h)[None, :]
        if within_ids:
            ix = list(map(self.lookup.__getitem__, within_ids))
            similarities = self.measure(h, self.h[ix, :])  # mypy: ignore
        else:
            similarities = self.measure(h, self.h)  # mypy: ignore
        similarities = similarities[0, :]
        logging.debug(similarities)
        scores = -numpy.sort(-similarities)
        ix = numpy.argsort(-similarities)[:n]
        ix = ix.tolist()
        scores = scores.tolist()
        _ids = [self.index[i] for i in ix]
        return _ids, scores

    def add(self, items: t.Sequence[VectorItem]) -> None:
        index = [item.id for item in items]
        h = numpy.stack([item.vector for item in items])

        if self.h is not None:
            old_not_in_new = list(set(self.index) - set(index))
            ix_old = [self.lookup[_id] for _id in old_not_in_new]
            h = numpy.concatenate((self.h[ix_old], h), axis=0)
            index = [self.index[i] for i in ix_old] + index

        return self._setup(h, index)

    def delete(self, ids):
        ix = list(map(self.lookup.__getitem__, ids))
        h = numpy.delete(self.h, ix, axis=0)
        index = [_id for _id in self.index if _id not in set(ids)]
        self._setup(h, index)



import dataclasses as dc
import importlib
import json
import os
import re
import tarfile
import typing as t

from aladindb.backends.local.artifacts import FileSystemArtifactStore
from aladindb.base.serializable import Serializable
from aladindb.components.serializer import serializers
from aladindb.misc.annotations import public_api
from aladindb.misc.download import Fetcher

from .component import Component

if t.TYPE_CHECKING:
    from aladindb.base.datalayer import Datalayer


ARTIFACTS = 'artifacts'
_STACK_JSON_FILE = 'stack.json'


@public_api(stability='alpha')
@dc.dataclass(kw_only=True)
class Stack(Component):
    """
    A placeholder to hold list of components under a namespace and packages them as
    a tarball
    This tarball can be retrieved back to a `Stack` instance with ``load`` method.
    {component_parameters}
    :param components: List of components to stack together and add to database.
    :param version: Version number of the model(?)
    """

    __doct__ = __doc__.format(component_parameters=Component.__doc__)

    type_id: t.ClassVar[str] = 'stack'

    components: t.Sequence[Component] = ()

    @property
    def child_components(self):
        return (('components', 'component_list'),)

    def _load_components(self):
        self._component_type_store = {}
        self._component_store = {}
        if self.components:
            for component in self.components:
                type_id = component.type_id
                identifier = component.identifier

                if not self._component_type_store.get(type_id, None):
                    self._component_type_store.update({type_id: [component]})
                else:
                    self._component_type_store[type_id] = self._component_type_store[
                        type_id
                    ].append(component)

                self._component_store.update({identifier: component})

    @staticmethod
    def to_tarball(folder_path: str):
        """
        Create a tarball (compressed archive) from a folder.

        :param folder_path: Path to the folder to be archived.
        """
        path = os.path.basename(folder_path) + '.tar.gz'
        with tarfile.open(path, "w:gz") as tar:
            tar.add(folder_path, arcname=os.path.basename(folder_path))

    @staticmethod
    def from_tarball(tarball_path: str):
        """
        Extract the contents of stack tarball

        :param tarball_path: Path to the tarball file.
        """
        extract_path = tarball_path.split('.tar.gz')[0]
        with tarfile.open(tarball_path, "r:gz") as tar:
            tar.extractall(path=extract_path)
        return extract_path

    def save(self, db: 'Datalayer', name: str):
        """
        Save method takes a stack name and exports ``self``
        to tarball by serializing ``Stack`` along with storing
        artifacts to filesystem

        The json serialized file and artifacts folders are packaged by tar
        ball.
        """
        os.makedirs(name, exist_ok=True)
        fs_store = FileSystemArtifactStore(os.path.join(name, ARTIFACTS))

        serialized, artifacts = self.serialized
        artifact_info = fs_store.save(artifacts)
        serialized = db.artifact_store.replace(serialized, artifact_info)

        with open(os.path.join(name, _STACK_JSON_FILE), 'w') as outfile:
            json.dump(serialized, outfile)

        self.to_tarball(name)

    def load(self, path: str):
        """
        Load method loads the stack tarball, loads the `stack.json` and `artifacts`.
        It binds the loaded stack object to ``self``

        Stack should be empty for it to load the given tarball else it will
        override the exisitng artibutes.

        :param: Path to tarball which contains artifacts and stack json.
        """
        # Extract tarball
        extracted_path = self.from_tarball(path + '.tar.gz')

        # Load stack.josn file
        with open(os.path.join(extracted_path, _STACK_JSON_FILE), 'r') as inputfile:
            serialized = json.load(inputfile)

        # Load the stack
        fs_store = os.path.join(extracted_path, ARTIFACTS)
        info = fs_store.load(serialized, lazy=True)
        loaded_stack = Serializable.deserialize(info)

        self.identifier = loaded_stack.identifier
        self.components = loaded_stack.components
        self._load_components()

    def get(self, identifier: str):
        return self._component_store.get(identifier, None)

    def get_component_type(self, type_id: str):
        return self._component_type_store.get(type_id, None)

    @staticmethod
    def from_dict(d: t.Dict, db: t.Optional['Datalayer'] = None):
        """
        Load a stack from a dictionary.

        :param d: Dictionary containing the stack definition.
        :param db: Datalayer instance to use for loading artifacts.

        The dictionary should look something like below.
        1. An ``identifier`` field
        2. An ``artifact`` field containing a list of artifacts including the
           serializer used
        3. A ``components`` field containing a list of component definitions.
           These definitions can recursively refer to other components or artifacts,
           resp. ``$components.<identifier>`` or ``$artifacts[<index>]``.

        >>> d = {
        ...    'identifier': 'my_stack',
        ...    'artifacts': [
        ...        {
        ...            'serializer': 'pickle',
        ...            'bytes': bytes_,
        ...        },
        ...    ],
        ...    'components': [
        ...        {'module': 'aladindb.ext.pillow', 'variable': 'pil_image'},
        ...        {
        ...            'module': 'aladindb.components.model',
        ...            'cls': 'Model',
        ...            'dict': {
        ...                'identifier': 'my_model',
        ...                'object': '$artifacts[0]',
        ...                'encoder': '$components.pil_image',
        ...            },
        ...        },
        ...    ],
        ... }
        >>> s = Stack.from_dict(d)
        >>> isinstance(s, Stack)
        True
        >>> isinstance(s.components[1], Model)
        True

        """
        return _load_stack(d, db)


def _find_variables(r):
    if 'dict' not in r:
        return []
    variables = []
    for k in r['dict']:
        if isinstance(r['dict'][k], str) and r['dict'][k].startswith('$components'):
            variables.append(r['dict'][k])
    return variables


def _find_leaves(components, existing):
    def contains_new_variable(r):
        if set(_find_variables(r)).issubset(set(existing)):
            return False
        return True

    out = [c for c in components if not contains_new_variable(c)]
    components = [c for c in components if contains_new_variable(c)]
    return out, components


def _load_artifact(variable, artifacts, serializers, dir=None, fetcher=None):
    serializer = serializers[artifacts[variable]['serializer']]
    if 'bytes' not in artifacts[variable]:
        fetcher = fetcher or Fetcher()
        try:
            uri = artifacts[variable]['uri']
        except KeyError:
            raise Exception(f'No uri or bytes found for artifact {variable}')

        artifacts[variable]['bytes'] = fetcher(uri)
    return serializer.decode(artifacts[variable]['bytes'])


def _build_component(d, artifacts, components, serializers):
    module = importlib.import_module(d['module'])
    if 'dict' in d:
        defn = d['dict'].copy()
        for k in defn:
            if isinstance(defn[k], str) and defn[k].startswith('$'):
                if '.' in defn[k]:
                    assert defn[k].startswith('$components.')
                    type, variable = defn[k][1:].split('.')
                else:
                    match = re.match('\$artifacts\[(\d+)\]', defn[k])
                    assert (
                        match
                    ), 'Unknown variable type {defn[k]}, expected $artifacts[<index>]'
                    type = 'artifacts'
                    variable = int(match.groups()[0].strip())

                if type == 'artifacts':
                    if 'artifact' not in artifacts[variable]:
                        artifacts[variable]['artifact'] = _load_artifact(
                            variable, artifacts, serializers=serializers
                        )
                    defn[k] = artifacts[variable]['artifact']
                elif type == 'components':
                    defn[k] = {c.identifier: c for c in components}[variable]
                else:
                    raise Exception(f'Unknown variable type {type}')

        component_cls = getattr(module, d['cls'])
        built = component_cls(**defn)
        if built.type_id == 'serializer':
            serializers[built.identifier] = built.object
    else:
        built = getattr(module, d['variable'])
    return built


def _load_stack(d, db=None):
    components = []
    to_build = d['components'].copy()
    while True:
        current, to_build = _find_leaves(
            to_build,
            existing=[f'$components.{c.identifier}' for c in components],
        )
        for c in current:
            built = _build_component(
                c,
                artifacts=d['artifacts'],
                components=components,
                serializers=db.serializers if db else serializers,
            )
            components.append(built)
        if not to_build:
            break
    return Stack(identifier=d['identifier'], components=components)



from __future__ import annotations

import dataclasses as dc
import typing as t
from functools import cached_property

import numpy
from overrides import override

from aladindb.backends.mongodb.query import Select
from aladindb.base.artifact import Artifact
from aladindb.base.datalayer import Datalayer
from aladindb.base.document import Document
from aladindb.components.component import Component
from aladindb.misc.annotations import public_api


@public_api(stability='stable')
@dc.dataclass(kw_only=True)
class Dataset(Component):
    """A dataset is an immutable collection of documents.
    {component_params}
    :param select: A query to select the documents for the dataset
    :param sample_size: The number of documents to sample from the query
    :param random_seed: The random seed to use for sampling
    :param creation_date: The date the dataset was created
    :param raw_data: The raw data for the dataset
    """

    __doc__ = __doc__.format(component_params=Component.__doc__)

    type_id: t.ClassVar[str] = 'dataset'

    select: t.Optional[Select] = None
    sample_size: t.Optional[int] = None
    random_seed: t.Optional[int] = None
    creation_date: t.Optional[str] = None
    raw_data: t.Optional[t.Union[Artifact, t.Any]] = None

    @override
    def pre_create(self, db: 'Datalayer') -> None:
        if self.raw_data is None:
            if self.select is None:
                raise ValueError('select cannot be None')
            data = list(db.execute(self.select))
            if self.sample_size is not None and self.sample_size < len(data):
                perm = self.random.permutation(len(data)).tolist()
                data = [data[perm[i]] for i in range(self.sample_size)]
            self.raw_data = Artifact(artifact=[r.encode() for r in data])

    @override
    def post_create(self, db: 'Datalayer') -> None:
        return self.on_load(db)

    @override
    def on_load(self, db: 'Datalayer') -> None:
        assert isinstance(self.raw_data, Artifact)
        self.data = [
            Document(Document.decode(r.copy(), encoders=db.encoders))
            for r in self.raw_data.artifact
        ]

    @cached_property
    def random(self):
        return numpy.random.default_rng(seed=self.random_seed)



"""
Operations on dictionaries used to fill and combine config files
and environment variables
"""

import os
import sys
import typing as t

Dict = t.Dict[str, object]
StrDict = t.Dict[str, str]

SEP = '_'
_NONE = object()


def combine_configs(dicts: t.Sequence[Dict]) -> Dict:
    result: Dict = {}
    for d in dicts:
        _combine_one(result, d)
    return result


def environ_to_config_dict(
    prefix: str,
    parent: StrDict,
    environ: t.Optional[StrDict] = None,
    err: t.Optional[t.TextIO] = sys.stderr,
    fail: bool = False,
):
    env_dict = _environ_dict(prefix, environ)
    good, bad = _env_dict_to_config_dict(env_dict, parent)

    if bad:
        bad = {k: ', '.join([prefix + i.upper() for i in v]) for k, v in bad.items()}
        msg = '\n'.join(f'{k}: {v}' for k, v in sorted(bad.items()))

        s = 's' * (sum(len(v) for v in bad.values()) != 1)
        msg = f'Bad environment variable{s}:\n{msg}'
        if err is not None:
            print(msg, file=err)
        if fail:
            raise ValueError(msg)

    return good


def _split_address(
    key: str, parent: StrDict
) -> t.Iterator[t.Tuple[StrDict, t.Tuple[str]]]:
    def split(key, parent, *address):
        if key in parent:
            yield *address, key

        for k, v in parent.items():
            if key.startswith(ks := k + SEP) and isinstance(v, dict):
                yield from split(key[len(ks) :], v, *address, k)

    return split(key, parent)


def _environ_dict(prefix: str, environ: t.Optional[StrDict] = None) -> StrDict:
    if not (prefix.isupper() and prefix.endswith(SEP) and not prefix.startswith(SEP)):
        raise ValueError(f'Bad prefix={prefix}')

    d = os.environ if environ is None else environ
    items = ((k, v) for k, v in d.items() if k.isupper() and k.startswith(prefix))
    return {k[len(prefix) :].lower(): v for k, v in items}


def _combine_one(target, source):
    for k, v in source.items():
        old_v = target.get(k, _NONE)
        if old_v is _NONE:
            target[k] = v

        elif (
            not (isinstance(v, type(old_v)) or isinstance(old_v, type(v)))
            and old_v is not None
        ):
            err = f'Expected {type(old_v)} but got {type(v)} for key={k}'
            raise ValueError(err)

        elif isinstance(v, dict):
            _combine_one(old_v, v)

        else:
            target[k] = v


def _env_dict_to_config_dict(
    env_dict: StrDict, parent: StrDict
) -> t.Tuple[StrDict, StrDict]:
    good: t.Dict = {}
    bad: t.Dict = {}

    for k, v in env_dict.items():
        addresses = list(_split_address(k, parent))
        if not addresses:
            bad.setdefault('unknown', []).append(k)
        elif len(addresses) > 1:
            bad.setdefault('ambiguous', []).append(k)
        else:
            d = good
            *address, last = addresses[0]
            for a in address:
                d = d.setdefault(a, {})
            d[last] = v

    return good, bad



import threading
import typing as t

Callback = t.Callable[[], None]


class Event(threading.Event):
    """
    A threading.Event that also calls back to zero or more functions when its state
    is set or reset, and has a __bool__ method.

    Note that the callback might happen on some completely different thread,
    so these functions cannot block"""

    on_set: t.List[Callback]

    def __init__(self, *on_set: Callback):
        self.on_set = list(on_set)
        super().__init__()

    def set(self):
        super().set()
        [c() for c in self.on_set]

    def clear(self):
        super().clear()
        [c() for c in self.on_set]

    def __bool__(self):
        return self.is_set()


class Runnable:
    """A base class for things that start, run, finish, stop and join

    Stopping is requesting immediate termination: finishing is saying that
    there is no more work to be done, finish what you are doing.

    A Runnable has two `Event`s, `running` and `stopped`, and you can either
    `wait` on either of these conditions to be true, or add a callback function
    (which must be non-blocking) to either of them.

    `running` is not set until the setup for a `Runnable` has finished;
    `stopped` is not set until all the computations in a thread have ceased.

    An Runnable can be used as a context manager:

        with runnable:
            # The runnable is running by this point
            do_stuff()
        # By the time you get to here, the runnable has completely stopped

    The above means roughly the same as

        runnable.start()
        try:
            do_stuff()
            runnable.finish()
            runnable.join()
        finally:
            runnable.stop()

    """

    #: An Event that is only set once this object is actually running
    running: Event

    #: An event that is only set once this object is fully stopped
    stopped: Event

    def __init__(self):
        self.running = Event()
        self.stopped = Event()

    def start(self):
        """Start this object.

        Note that self.running might not be immediately true after this method completes
        """
        self.running.set()

    def stop(self):
        """Stop as soon as possible. might not do anything, should never raise.

        Note that self.stopped might not be immediately true after this method completes
        """
        self.running.clear()
        self.stopped.set()

    def finish(self):
        """Request an orderly shutdown where all existing work is completed.

        Note that self.stopped might not be immediately true after this method completes
        """
        self.stop()

    def join(self, timeout: t.Optional[float] = None):
        """Join this thread or process.  Might block indefinitely, might do nothing"""

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            if exc_type is None:
                self.finish()
                self.join()
        finally:
            self.stop()



import typing as t

import click
import tenacity
from pymongo.results import DeleteResult, InsertOneResult, UpdateResult

from aladindb import logging
from aladindb.backends.base.metadata import MetaDataStore
from aladindb.components.component import Component
from aladindb.misc.colors import Colors


class MongoMetaDataStore(MetaDataStore):
    """
    Metadata store for MongoDB.

    :param conn: MongoDB client connection
    :param name: Name of database to host filesystem
    """

    def __init__(
        self,
        conn: t.Any,
        name: t.Optional[str] = None,
    ) -> None:
        self.name = name
        self.db = conn[name]
        self.meta_collection = self.db['_meta']
        self.cdc_collection = self.db['_cdc_tables']
        self.component_collection = self.db['_objects']
        self.job_collection = self.db['_jobs']
        self.parent_child_mappings = self.db['_parent_child_mappings']
        self.conn = conn

    def url(self):
        return self.conn.HOST + ':' + str(self.conn.PORT) + '/' + self.name

    def drop(self, force: bool = False):
        if not force:
            if not click.confirm(
                f'{Colors.RED}[!!!WARNING USE WITH CAUTION AS YOU '
                f'WILL LOSE ALL DATA!!!]{Colors.RESET} '
                'Are you sure you want to drop all meta-data? ',
                default=False,
            ):
                logging.warn('Aborting...')
        self.db.drop_collection(self.meta_collection.name)
        self.db.drop_collection(self.component_collection.name)
        self.db.drop_collection(self.job_collection.name)
        self.db.drop_collection(self.parent_child_mappings.name)

    def create_parent_child(self, parent: str, child: str) -> None:
        self.parent_child_mappings.insert_one(
            {
                'parent': parent,
                'child': child,
            }
        )

    def create_component(self, info: t.Dict) -> InsertOneResult:
        if 'hidden' not in info:
            info['hidden'] = False
        return self.component_collection.insert_one(info)

    def create_job(self, info: t.Dict) -> InsertOneResult:
        return self.job_collection.insert_one(info)

    def get_parent_child_relations(self):
        c = self.parent_child_mappings.find()
        return [(r['parent'], r['child']) for r in c]

    def get_component_version_children(self, unique_id: str):
        return self.parent_child_mappings.distinct('child', {'parent': unique_id})

    def get_job(self, identifier: str):
        return self.job_collection.find_one({'identifier': identifier})

    def create_metadata(self, key: str, value: str):
        return self.meta_collection.insert_one({'key': key, 'value': value})

    def get_metadata(self, key: str):
        return self.meta_collection.find_one({'key': key})['value']

    def update_metadata(self, key: str, value: str):
        return self.meta_collection.update_one({'key': key}, {'$set': {'value': value}})

    def get_latest_version(
        self, type_id: str, identifier: str, allow_hidden: bool = False
    ) -> int:
        try:
            if allow_hidden:
                return sorted(
                    self.component_collection.distinct(
                        'version', {'identifier': identifier, 'type_id': type_id}
                    )
                )[-1]
            else:
                return sorted(
                    self.component_collection.distinct(
                        'version',
                        {
                            '$or': [
                                {
                                    'identifier': identifier,
                                    'type_id': type_id,
                                    'hidden': False,
                                },
                                {
                                    'identifier': identifier,
                                    'type_id': type_id,
                                    'hidden': {'$exists': 0},
                                },
                            ]
                        },
                    )
                )[-1]
        except IndexError:
            raise FileNotFoundError(f'Can\'t find {type_id}: {identifier} in metadata')

    def update_job(self, identifier: str, key: str, value: t.Any) -> UpdateResult:
        return self.job_collection.update_one(
            {'identifier': identifier}, {'$set': {key: value}}
        )

    def show_components(self, type_id: str, **kwargs) -> t.List[t.Union[t.Any, str]]:
        # TODO: Should this be sorted?
        return self.component_collection.distinct(
            'identifier', {'type_id': type_id, **kwargs}
        )

    # TODO: Why is this is needed to prevent failures in CI?
    @tenacity.retry(stop=tenacity.stop_after_attempt(10))
    def show_component_versions(
        self, type_id: str, identifier: str
    ) -> t.List[t.Union[t.Any, int]]:
        return self.component_collection.distinct(
            'version', {'type_id': type_id, 'identifier': identifier}
        )

    def list_components_in_scope(self, scope: str):
        out = []
        for r in self.component_collection.find({'parent': scope}):
            out.append((r['type_id'], r['identifier']))
        return out

    def show_job(self, job_id: str):
        return self.job_collection.find_one({'identifier': job_id})

    def show_jobs(self, status=None):
        status = {} if status is None else {'status': status}
        return list(
            self.job_collection.find(
                status, {'identifier': 1, '_id': 0, 'method': 1, 'status': 1, 'time': 1}
            )
        )

    def _component_used(
        self, type_id: str, identifier: str, version: t.Optional[int] = None
    ) -> bool:
        if version is None:
            members: t.Union[t.Dict, str] = {'$regex': f'^{identifier}/{type_id}'}
        else:
            members = Component.make_unique_id(type_id, identifier, version)

        return bool(self.component_collection.count_documents({'members': members}))

    def component_has_parents(self, type_id: str, identifier: str) -> int:
        doc = {'child': {'$regex': f'^{type_id}/{identifier}/'}}
        return self.parent_child_mappings.count_documents(doc)

    def component_version_has_parents(
        self, type_id: str, identifier: str, version: int
    ) -> int:
        doc = {'child': Component.make_unique_id(type_id, identifier, version)}
        return self.parent_child_mappings.count_documents(doc)

    def delete_component_version(
        self, type_id: str, identifier: str, version: int
    ) -> DeleteResult:
        if self._component_used(type_id, identifier, version=version):
            raise Exception('Component version already in use in other components!')

        self.parent_child_mappings.delete_many(
            {'parent': Component.make_unique_id(type_id, identifier, version)}
        )

        return self.component_collection.delete_many(
            {
                'identifier': identifier,
                'type_id': type_id,
                'version': version,
            }
        )

    def _get_component(
        self,
        type_id: str,
        identifier: str,
        version: int,
        allow_hidden: bool = False,
    ) -> t.Dict[str, t.Any]:
        if not allow_hidden:
            r = self.component_collection.find_one(
                {
                    '$or': [
                        {
                            'identifier': identifier,
                            'type_id': type_id,
                            'version': version,
                            'hidden': False,
                        },
                        {
                            'identifier': identifier,
                            'type_id': type_id,
                            'version': version,
                            'hidden': {'$exists': 0},
                        },
                    ]
                }
            )
        else:
            r = self.component_collection.find_one(
                {
                    'identifier': identifier,
                    'type_id': type_id,
                    'version': version,
                },
            )
        return r

    def get_component_version_parents(self, unique_id: str) -> t.List[str]:
        return [
            r['parent'] for r in self.parent_child_mappings.find({'child': unique_id})
        ]

    def _replace_object(
        self,
        info: t.Dict[str, t.Any],
        identifier: str,
        type_id: str,
        version: int,
    ) -> None:
        self.component_collection.replace_one(
            {'identifier': identifier, 'type_id': type_id, 'version': version},
            info,
        )

    def _update_object(
        self,
        identifier: str,
        type_id: str,
        key: str,
        value: t.Any,
        version: int,
    ):
        return self.component_collection.update_one(
            {'identifier': identifier, 'type_id': type_id, 'version': version},
            {'$set': {key: value}},
        )

    def write_output_to_job(self, identifier, msg, stream):
        if stream not in ('stdout', 'stderr'):
            raise ValueError(f'stream is "{stream}", should be stdout or stderr')
        self.job_collection.update_one(
            {'identifier': identifier}, {'$push': {stream: msg}}
        )

    def hide_component_version(
        self, type_id: str, identifier: str, version: int
    ) -> None:
        self.component_collection.update_one(
            {'type_id': type_id, 'identifier': identifier, 'version': version},
            {'$set': {'hidden': True}},
        )

    def disconnect(self):
        """
        Disconnect the client
        """

        # TODO: implement me



from __future__ import annotations

import dataclasses as dc
import multiprocessing
import typing as t
from abc import abstractmethod
from functools import wraps

import tqdm
from overrides import override
from sklearn.pipeline import Pipeline

from aladindb import logging
from aladindb.backends.base.metadata import NonExistentMetadataError
from aladindb.backends.base.query import CompoundSelect, Select, TableOrCollection
from aladindb.backends.ibis.field_types import FieldType
from aladindb.backends.ibis.query import IbisCompoundSelect, Table
from aladindb.backends.query_dataset import QueryDataset
from aladindb.base.artifact import Artifact
from aladindb.base.serializable import Serializable
from aladindb.components.component import Component
from aladindb.components.encoder import Encoder
from aladindb.components.metric import Metric
from aladindb.components.schema import Schema
from aladindb.jobs.job import ComponentJob, Job
from aladindb.misc.annotations import public_api
from aladindb.misc.special_dicts import MongoStyleDict

if t.TYPE_CHECKING:
    from aladindb.base.datalayer import Datalayer
    from aladindb.components.dataset import Dataset

EncoderArg = t.Union[Encoder, FieldType, str, None]
ObjectsArg = t.Sequence[t.Union[t.Any, Artifact]]


class _to_call:
    def __init__(self, callable, **kwargs):
        self.callable = callable
        self.kwargs = kwargs

    def __call__(self, X):
        return self.callable(X, **self.kwargs)


@dc.dataclass(kw_only=True)
class _TrainingConfiguration(Component):
    """
    Training configuration object, containing all settings necessary for a particular
    learning-task use-case to be serialized and initiated. The object is ``callable``
    and returns a class which may be invoked to apply training.

    :param **kwargs: Key-values pairs, the variables which configure training.
    """

    kwargs: t.Optional[t.Dict] = None
    type_id: t.ClassVar[str] = 'training_configuration'

    def get(self, k, default=None):
        try:
            return getattr(self, k)
        except AttributeError:
            return self.kwargs.get(k, default)


@dc.dataclass(kw_only=True)
class _Predictor:
    # Mixin class for components which can predict.
    """:param encoder: Encoder instance
    :param output_schema: Output schema (mapping of encoders)
    :param flatten: Flatten the model outputs
    :param preprocess: Preprocess function
    :param postprocess: Postprocess function
    :param collate_fn: Collate function
    :param batch_predict: Whether to batch predict
    :param takes_context: Whether the model takes context into account
    :param metrics: The metrics to evaluate on
    :param model_update_kwargs: The kwargs to use for model update
    :param validation_sets: The validation ``Dataset`` instances to use
    :param predict_X: The key of the input data to use for .predict
    :param predict_select: The select to use for .predict
    :param predict_max_chunk_size: The max chunk size to use for .predict
    :param predict_kwargs: The kwargs to use for .predict"""

    type_id: t.ClassVar[str] = 'model'

    encoder: EncoderArg = None
    output_schema: t.Optional[t.Union[Schema, dict]] = None
    flatten: bool = False
    preprocess: t.Union[t.Callable, Artifact, None] = None
    postprocess: t.Union[t.Callable, Artifact, None] = None
    collate_fn: t.Union[t.Callable, Artifact, None] = None
    batch_predict: bool = False
    takes_context: bool = False
    metrics: t.Sequence[t.Union[str, Metric, None]] = ()
    model_update_kwargs: t.Dict = dc.field(default_factory=dict)
    validation_sets: t.Optional[t.Sequence[t.Union[str, Dataset]]] = None

    predict_X: t.Optional[str] = None
    predict_select: t.Optional[CompoundSelect] = None
    predict_max_chunk_size: t.Optional[int] = None
    predict_kwargs: t.Optional[t.Dict] = None

    @abstractmethod
    def to_call(self, X, *args, **kwargs):
        """
        The method to use to call prediction. Should be implemented
        by the child class.
        """

    def create_predict_job(
        self,
        X: str,
        select: t.Optional[Select] = None,
        ids: t.Optional[t.Sequence[str]] = None,
        max_chunk_size: t.Optional[int] = None,
        **kwargs,
    ):
        return ComponentJob(
            component_identifier=self.identifier,
            method_name='predict',
            type_id='model',
            args=[X],
            kwargs={
                'select': select.serialize() if select else None,
                'ids': ids,
                'max_chunk_size': max_chunk_size,
                **kwargs,
            },
        )

    async def _apredict_one(self, X: t.Any, **kwargs):
        raise NotImplementedError

    async def _apredict(self, X: t.Any, one: bool = False, **kwargs):
        raise NotImplementedError

    def _predict_one(self, X: t.Any, **kwargs) -> int:
        if isinstance(self.preprocess, Artifact):
            X = self.preprocess.artifact(X)
        elif self.preprocess is not None:
            raise ValueError('Bad preprocess')

        output = self.to_call(X, **kwargs)

        if isinstance(self.postprocess, Artifact):
            output = self.postprocess.artifact(output)
        elif self.postprocess is not None:
            raise ValueError('Bad postprocess')

        return output

    def _forward(
        self, X: t.Sequence[int], num_workers: int = 0, **kwargs
    ) -> t.Sequence[int]:
        if self.batch_predict:
            return self.to_call(X, **kwargs)

        outputs = []
        if num_workers:
            to_call = _to_call(self.to_call, **kwargs)
            pool = multiprocessing.Pool(processes=num_workers)
            for r in pool.map(to_call, X):
                outputs.append(r)
            pool.close()
            pool.join()
        else:
            for r in X:
                outputs.append(self.to_call(r, **kwargs))
        return outputs

    def _predict(self, X: t.Any, one: bool = False, **predict_kwargs):
        if one:
            return self._predict_one(X)

        if isinstance(self.preprocess, Artifact):
            X = [self.preprocess.artifact(i) for i in X]
        elif self.preprocess is not None:
            raise ValueError('Bad preprocess')

        if isinstance(self.collate_fn, Artifact):
            raise ValueError('Bad collate function')

        elif self.collate_fn is not None:
            X = self.collate_fn(X)

        outputs = self._forward(X, **predict_kwargs)

        if isinstance(self.postprocess, Artifact):
            outputs = [self.postprocess.artifact(o) for o in outputs]
        elif self.postprocess is not None:
            raise ValueError('Bad postprocess')

        return outputs

    def predict(
        self,
        X: t.Any,
        db: t.Optional[Datalayer] = None,
        select: t.Optional[CompoundSelect] = None,
        ids: t.Optional[t.List[str]] = None,
        max_chunk_size: t.Optional[int] = None,
        dependencies: t.Sequence[Job] = (),
        listen: bool = False,
        one: bool = False,
        context: t.Optional[t.Dict] = None,
        insert_to: t.Optional[t.Union[TableOrCollection, str]] = None,
        key: t.Optional[str] = None,
        in_memory: bool = True,
        overwrite: bool = False,
        **kwargs,
    ) -> t.Any:
        was_added = self.db is not None
        db = self.db or db

        if one:
            assert select is None, 'select must be None when ``one=True`` (direct call)'

        if isinstance(select, dict):
            select = Serializable.deserialize(select)

        if isinstance(select, Table):
            select = select.to_query()

        if db is not None:
            if isinstance(select, IbisCompoundSelect):
                from aladindb.backends.sqlalchemy.metadata import SQLAlchemyMetadata

                assert isinstance(db.metadata, SQLAlchemyMetadata)
                try:
                    _ = db.metadata.get_query(str(hash(select)))
                except NonExistentMetadataError:
                    logging.info(f'Query {select} not found in metadata, adding...')
                    db.metadata.add_query(select, self.identifier)
                    logging.info('Done')

            if not was_added:
                logging.info(f'Adding model {self.identifier} to db')
                assert isinstance(self, Component)
                db.add(self)

        if listen:
            assert db is not None
            assert select is not None
            return self._predict_and_listen(
                X=X,
                db=db,
                select=select,
                max_chunk_size=max_chunk_size,
                **kwargs,
            )

        # TODO: tidy up this logic
        if select is not None and db is not None and db.compute.type == 'distributed':
            return self.create_predict_job(
                X,
                select=select,
                ids=ids,
                max_chunk_size=max_chunk_size,
                overwrite=overwrite,
                **kwargs,
            )(db=db, dependencies=dependencies)
        else:
            if select is not None and ids is None:
                assert db is not None
                return self._predict_with_select(
                    X=X,
                    select=select,
                    db=db,
                    in_memory=in_memory,
                    max_chunk_size=max_chunk_size,
                    overwrite=overwrite,
                    **kwargs,
                )
            elif select is not None and ids is not None:
                assert db is not None
                return self._predict_with_select_and_ids(
                    X=X,
                    select=select,
                    ids=ids,
                    db=db,
                    max_chunk_size=max_chunk_size,
                    in_memory=in_memory,
                    **kwargs,
                )
            else:
                if self.takes_context:
                    kwargs['context'] = context

                output = self._predict(
                    (X[key] if one else [r[key] for r in X]) if key else X,
                    one=one,
                    **kwargs,
                )
                if insert_to is not None:
                    msg = (
                        '`self.db` has not been set; this is necessary if'
                        ' `insert_to` is not None; use `db.add(self)`'
                    )

                    from aladindb.base.datalayer import Datalayer

                    assert isinstance(db, Datalayer), msg
                    if isinstance(insert_to, str):
                        insert_to = db.load(
                            'table',
                            insert_to,
                        )  # type: ignore[assignment]
                    if one:
                        output = [output]

                    assert isinstance(insert_to, TableOrCollection)
                    ids, _ = db.execute(insert_to.insert([X] if one else X))
                    ids = t.cast(t.List[t.Any], ids)
                    assert isinstance(key, str)

                    insert_to.model_update(
                        db=db,
                        model=self.identifier,
                        outputs=output,
                        key=key,
                        version=self.version,
                        ids=ids,
                        flatten=self.flatten,
                        **self.model_update_kwargs,
                    )
                return output

    async def apredict(
        self,
        X: t.Any,
        context: t.Optional[t.Dict] = None,
        one: bool = False,
        **kwargs,
    ):
        if self.takes_context:
            kwargs['context'] = context
        return await self._apredict(X, one=one, **kwargs)

    def _predict_and_listen(
        self,
        X: t.Any,
        select: CompoundSelect,
        db: Datalayer,
        in_memory: bool = True,
        max_chunk_size: t.Optional[int] = None,
        dependencies: t.Sequence[Job] = (),
        **kwargs,
    ):
        from aladindb.components.listener import Listener

        return db.add(
            Listener(
                key=X,
                model=t.cast(Model, self),
                select=select,
                predict_kwargs={
                    **kwargs,
                    'in_memory': in_memory,
                    'max_chunk_size': max_chunk_size,
                },
            ),
            dependencies=dependencies,
        )[0]

    def _predict_with_select(
        self,
        X: t.Any,
        select: Select,
        db: Datalayer,
        max_chunk_size: t.Optional[int] = None,
        in_memory: bool = True,
        overwrite: bool = False,
        **kwargs,
    ):
        ids = []
        if not overwrite:
            query = select.select_ids_of_missing_outputs(
                key=X,
                model=self.identifier,
                version=t.cast(int, self.version),
            )
        else:
            query = select.select_ids

        try:
            id_field = db.databackend.id_field
        except AttributeError:
            id_field = query.table_or_collection.primary_id

        for r in tqdm.tqdm(db.execute(query)):
            ids.append(str(r[id_field]))

        return self._predict_with_select_and_ids(
            X=X,
            db=db,
            ids=ids,
            select=select,
            max_chunk_size=max_chunk_size,
            in_memory=in_memory,
            **kwargs,
        )

    def _predict_with_select_and_ids(
        self,
        X: t.Any,
        db: Datalayer,
        select: Select,
        ids: t.List[str],
        in_memory: bool = True,
        max_chunk_size: t.Optional[int] = None,
        **kwargs,
    ):
        if max_chunk_size is not None:
            it = 0
            for i in range(0, len(ids), max_chunk_size):
                logging.info(f'Computing chunk {it}/{int(len(ids) / max_chunk_size)}')
                self._predict_with_select_and_ids(
                    X=X,
                    db=db,
                    ids=ids[i : i + max_chunk_size],
                    select=select,
                    max_chunk_size=None,
                    in_memory=in_memory,
                    **kwargs,
                )
                it += 1
            return

        X_data: t.Any
        if in_memory:
            if db is None:
                raise ValueError('db cannot be None')
            docs = list(db.execute(select.select_using_ids(ids)))
            if X != '_base':
                X_data = [MongoStyleDict(r.unpack())[X] for r in docs]
            else:
                X_data = [r.unpack() for r in docs]
        else:
            X_data = QueryDataset(
                select=select,
                ids=ids,
                fold=None,
                db=db,
                in_memory=False,
                keys=[X],
            )

        if len(X_data) > len(ids):
            raise Exception(
                'You\'ve specified more documents than unique ids;'
                f' Is it possible that {select.table_or_collection.primary_id}'
                f' isn\'t unique identifying?'
            )

        outputs = self.predict(X=X_data, one=False, **kwargs)

        if isinstance(self.encoder, Encoder):
            if self.flatten:
                outputs = [
                    [self.encoder(x).encode() for x in output] for output in outputs
                ]
            else:
                outputs = [self.encoder(x).encode() for x in outputs]
        elif isinstance(self.output_schema, Schema):
            encoded_ouputs = []
            for output in outputs:
                if isinstance(output, dict):
                    encoded_ouputs.append(self.output_schema.encode(output))
                elif self.flatten:
                    encoded_output = [self.output_schema.encode(x) for x in output]
                    encoded_ouputs.append(encoded_output)
            outputs = encoded_ouputs if encoded_ouputs else outputs

        assert isinstance(self.version, int)

        logging.info(f'Adding {len(outputs)} model outputs to `db`')
        select.model_update(
            db=db,
            model=self.identifier,
            outputs=outputs,
            key=X,
            version=self.version,
            ids=ids,
            flatten=self.flatten,
            **self.model_update_kwargs,
        )


@public_api(stability='stable')
@dc.dataclass(kw_only=True)
class Model(_Predictor, Component):
    """Model component which wraps a model to become serializable
    {component_params}
    {_predictor_params}
    :param object: Model object, e.g. sklearn model, etc..
    :param model_to_device_method: The method to transfer the model to a device
    :param metric_values: The metric values
    :param predict_method: The method to use for prediction
    :param model_update_kwargs: The kwargs to use for model update
    :param serializer: Serializer to store model to artifact store
    :param device: The device to use
    :param preferred_devices: The preferred devices to use
    :param training_configuration: The training configuration
    :param train_X: The key of the input data to use for training
    :param train_y: The key of the target data to use for training
    :param train_select: The select to use for training
    """

    __doc__ = __doc__.format(
        component_params=Component.__doc__,
        _predictor_params=_Predictor.__doc__,
    )

    object: t.Union[Artifact, t.Any, None]
    model_to_device_method: t.Optional[str] = None
    metric_values: t.Optional[t.Dict] = dc.field(default_factory=dict)
    predict_method: t.Optional[str] = None
    model_update_kwargs: dict = dc.field(default_factory=dict)
    serializer: str = 'dill'
    device: str = "cpu"
    preferred_devices: t.Union[None, t.Sequence[str]] = ("cuda", "mps", "cpu")

    training_configuration: t.Union[str, _TrainingConfiguration, None] = None
    train_X: t.Optional[str] = None
    train_y: t.Optional[str] = None
    train_select: t.Optional[CompoundSelect] = None

    artifact_attributes: t.ClassVar[t.Sequence[str]] = (
        'object',
        'preprocess',
        'postprocess',
    )
    type_id: t.ClassVar[str] = 'model'

    def __post_init__(self):
        super().__post_init__()

        if not isinstance(self.object, Artifact) and self.object is not None:
            self.object = Artifact(artifact=self.object, serializer=self.serializer)
        if self.preprocess and not isinstance(self.preprocess, Artifact):
            self.preprocess = Artifact(artifact=self.preprocess)
        if self.postprocess and not isinstance(self.postprocess, Artifact):
            self.postprocess = Artifact(artifact=self.postprocess)

        self._artifact_method = None
        if self.model_to_device_method is not None:
            self._artifact_method = getattr(self, self.model_to_device_method)

    def to_call(self, X, *args, **kwargs):
        if self.predict_method is None:
            return self.object.artifact(X, *args, **kwargs)
        out = getattr(self.object.artifact, self.predict_method)(X, *args, **kwargs)
        return out

    def post_create(self, db: Datalayer) -> None:
        if isinstance(self.training_configuration, str):
            self.training_configuration = db.load(
                'training_configuration', self.training_configuration
            )  # type: ignore[assignment]
        if isinstance(self.output_schema, Schema):
            db.add(self.output_schema)
        output_component = db.databackend.create_model_table_or_collection(self)
        if output_component is not None:
            db.add(output_component)

    # TODO - bring inside post_create
    @override
    def schedule_jobs(
        self,
        db: Datalayer,
        dependencies: t.Sequence[Job] = (),
        verbose: bool = False,
    ) -> t.Sequence[t.Any]:
        jobs = []
        if self.train_X is not None:
            assert (
                isinstance(self.training_configuration, _TrainingConfiguration)
                or self.training_configuration is None
            )
            assert self.train_select is not None
            jobs.append(
                self.fit(
                    X=self.train_X,
                    y=self.train_y,
                    configuration=self.training_configuration,
                    select=self.train_select,
                    db=db,
                    dependencies=dependencies,
                    metrics=self.metrics,  # type: ignore[arg-type]
                    validation_sets=self.validation_sets,
                )
            )
        if self.predict_X is not None:
            assert self.predict_select is not None
            jobs.append(
                self.predict(
                    X=self.predict_X,
                    select=self.predict_select,
                    max_chunk_size=self.predict_max_chunk_size,
                    db=db,
                    **(self.predict_kwargs or {}),
                )
            )
        return jobs

    def on_load(self, db: Datalayer) -> None:
        logging.debug(f'Calling on_load method of {self}')
        if self._artifact_method and self.preferred_devices:
            for i, device in enumerate(self.preferred_devices):
                try:
                    self._artifact_method(device)
                    self.device = device
                    return
                except Exception:
                    if i == len(self.preferred_devices) - 1:
                        raise

    @property
    def child_components(self) -> t.Sequence[t.Tuple[str, str]]:
        out = []
        if isinstance(self.encoder, Encoder):
            out.append(('encoder', 'encoder'))
        if self.training_configuration is not None:
            out.append(('training_configuration', 'training_configuration'))
        return out

    @property
    def training_keys(self) -> t.List:
        if isinstance(self.train_X, list):
            out = list(self.train_X)
        elif self.train_X is not None:
            out = [self.train_X]
        if self.train_y is not None:
            if isinstance(self.train_y, list):
                out.extend(self.train_y)
            else:
                out.append(self.train_y)
        return out

    def append_metrics(self, d: t.Dict[str, float]) -> None:
        if self.metric_values is not None:
            for k, v in d.items():
                self.metric_values.setdefault(k, []).append(v)

    def validate(
        self, db, validation_set: t.Union[Dataset, str], metrics: t.Sequence[Metric]
    ):
        db.add(self)
        out = self._validate(db, validation_set, metrics)
        if self.metric_values is None:
            raise ValueError('self.metric_values cannot be None')
        self.metric_values.update(out)
        db.metadata.update_object(
            type_id='model',
            identifier=self.identifier,
            version=self.version,
            key='dict.metric_values',
            value=self.metric_values,
        )

    def pre_create(self, db: Datalayer):
        if isinstance(self.encoder, str):
            self.encoder = db.load('encoder', self.encoder)  # type: ignore[assignment]

    def _validate(
        self,
        db: Datalayer,
        validation_set: t.Union[Dataset, str],
        metrics: t.Sequence[Metric],
    ):
        if isinstance(validation_set, str):
            from aladindb.components.dataset import Dataset

            validation_set = t.cast(Dataset, db.load('dataset', validation_set))

        mdicts = [MongoStyleDict(r.unpack()) for r in validation_set.data]
        assert self.train_X is not None
        prediction = self._predict([d[self.train_X] for d in mdicts])
        assert self.train_y is not None
        target = [d[self.train_y] for d in mdicts]
        assert isinstance(prediction, list)
        assert isinstance(target, list)
        results = {}

        for m in metrics:
            out = m(prediction, target)
            results[f'{validation_set.identifier}/{m.identifier}'] = out
        return results

    def create_fit_job(
        self,
        X: t.Union[str, t.Sequence[str]],
        select: t.Optional[Select] = None,
        y: t.Optional[str] = None,
        **kwargs,
    ):
        return ComponentJob(
            component_identifier=self.identifier,
            method_name='fit',
            type_id='model',
            args=[X],
            kwargs={
                'y': y,
                'select': select.serialize() if select else None,
                **kwargs,
            },
        )

    @abstractmethod
    def _fit(
        self,
        X: t.Any,
        y: t.Optional[t.Any] = None,
        configuration: t.Optional[_TrainingConfiguration] = None,
        data_prefetch: bool = False,
        db: t.Optional[Datalayer] = None,
        metrics: t.Optional[t.Sequence[Metric]] = None,
        select: t.Optional[Select] = None,
        validation_sets: t.Optional[t.Sequence[t.Union[str, Dataset]]] = None,
    ):
        pass

    def fit(
        self,
        X: t.Any,
        y: t.Optional[t.Any] = None,
        configuration: t.Optional[_TrainingConfiguration] = None,
        data_prefetch: bool = False,
        db: t.Optional[Datalayer] = None,
        dependencies: t.Sequence[Job] = (),
        metrics: t.Optional[t.Sequence[Metric]] = None,
        select: t.Optional[Select] = None,
        validation_sets: t.Optional[t.Sequence[t.Union[str, Dataset]]] = None,
        **kwargs,
    ) -> t.Optional[Pipeline]:
        """
        Fit the model on the given data.

        :param X: The key of the input data to use for training
        :param y: The key of the target data to use for training
        :param configuration: The training configuration (optional)
        :param data_prefetch: Whether to prefetch the data (optional)
        :param db: The datalayer (optional)
        :param dependencies: The dependencies (optional)
        :param metrics: The metrics to evaluate on (optional)
        :param select: The select to use for training (optional)
        :param validation_sets: The validation ``Dataset`` instances to use (optional)
        """
        if isinstance(select, dict):
            select = Serializable.deserialize(select)

        if validation_sets:
            from aladindb.components.dataset import Dataset

            validation_sets = list(validation_sets)
            for i, vs in enumerate(validation_sets):
                if isinstance(vs, Dataset):
                    assert db is not None
                    db.add(vs)
                    validation_sets[i] = vs.identifier

        if db is not None:
            db.add(self)

        if db is not None and db.compute.type == 'distributed':
            return self.create_fit_job(
                X,
                select=select,
                y=y,
                **kwargs,
            )(db=db, dependencies=dependencies)
        else:
            return self._fit(
                X,
                y=y,
                configuration=configuration,
                data_prefetch=data_prefetch,
                db=db,
                metrics=metrics,
                select=select,
                validation_sets=validation_sets,
                **kwargs,
            )


@wraps(_TrainingConfiguration)
def TrainingConfiguration(identifier: str, **kwargs):
    return _TrainingConfiguration(identifier=identifier, kwargs=kwargs)


@public_api(stability='beta')
@dc.dataclass(kw_only=True)
class APIModel(Component, _Predictor):
    '''{component_params}
    {predictor_params}
    :param model: The model to use, e.g. ``'text-embedding-ada-002'``'''

    __doc__ = __doc__.format(
        component_params=Component.__doc__,
        predictor_params=_Predictor.__doc__,
    )

    model: t.Optional[str] = None

    def __post_init__(self):
        super().__post_init__()
        if self.model is None:
            assert self.identifier is not None
            self.model = self.identifier

    def post_create(self, db: Datalayer) -> None:
        # TODO: This not necessary since added as a subcomponent
        if isinstance(self.output_schema, Schema):
            db.add(self.output_schema)
        # TODO add this logic to pre_create,
        # since then the `.add` clause is not necessary
        output_component = db.databackend.create_model_table_or_collection(self)
        if output_component is not None:
            db.add(output_component)

    @override
    def schedule_jobs(
        self,
        db: Datalayer,
        dependencies: t.Sequence[Job] = (),
        verbose: bool = False,
    ) -> t.Sequence[t.Any]:
        jobs = []
        if self.predict_X is not None:
            assert self.predict_select is not None
            jobs.append(
                self.predict(
                    X=self.predict_X,
                    select=self.predict_select,
                    max_chunk_size=self.predict_max_chunk_size,
                    db=db,
                    **(self.predict_kwargs or {}),
                )
            )
        return jobs

    @property
    def child_components(self):
        if isinstance(self.encoder, Encoder):
            return [('encoder', 'encoder')]
        return []


@dc.dataclass(kw_only=True)
class QueryModel(Component, _Predictor):
    """
    Model which can be used to query data and return those
    results as pre-computed queries.

    :param select: query used to find data (can include `like`)
    :param postprocess: postprocess function
    """

    select: CompoundSelect
    postprocess: t.Optional[t.Union[t.Callable, Artifact]] = None

    def __post_init__(self):
        super().__post_init__()
        if not isinstance(self.postprocess, Artifact):
            self.postprocess = Artifact(artifact=self.postprocess)

    @override
    def schedule_jobs(
        self,
        db: Datalayer,
        dependencies: t.Sequence[Job] = (),
        verbose: bool = False,
    ) -> t.Sequence[t.Any]:
        jobs = []
        if self.predict_X is not None:
            assert self.predict_select is not None
            jobs.append(
                self.predict(
                    X=self.predict_X,
                    select=self.predict_select,
                    max_chunk_size=self.predict_max_chunk_size,
                    db=db,
                    **(self.predict_kwargs or {}),
                )
            )
        return jobs

    def _predict_one(self, X: t.Any, **kwargs):
        select = self.select.set_variables(db=self.db, X=X)
        out = self.db.execute(select)
        if self.postprocess is not None:
            assert isinstance(self.postprocess, Artifact)
            return self.postprocess.artifact(out)
        return out

    def _predict(self, X: t.Any, one: bool = False, **predict_kwargs):
        if one:
            return self._predict_one(X, **predict_kwargs)
        return [self._predict_one(x, **predict_kwargs) for x in X]


@dc.dataclass(kw_only=True)
class SequentialModel(Component, _Predictor):
    """
    Sequential model component which wraps a model to become serializable

    {component_params}
    {_predictor_params}
    :param predictors: A list of predictors to use
    """

    __doc__ = __doc__.format(
        component_params=Component.__doc__,
        _predictor_params=_Predictor.__doc__,
    )
    predictors: t.List[t.Union[str, Model, APIModel]]

    @override
    def schedule_jobs(
        self,
        db: Datalayer,
        dependencies: t.Sequence[Job] = (),
        verbose: bool = False,
    ) -> t.Sequence[t.Any]:
        jobs = []
        if self.predict_X is not None:
            assert self.predict_select is not None
            jobs.append(
                self.predict(
                    X=self.predict_X,
                    select=self.predict_select,
                    max_chunk_size=self.predict_max_chunk_size,
                    db=db,
                    **(self.predict_kwargs or {}),
                )
            )
        return jobs

    def post_create(self, db: Datalayer):
        for p in self.predictors:
            if isinstance(p, str):
                continue
            p.post_create(db)
        self.on_load(db)

    def on_load(self, db: Datalayer):
        for i, p in enumerate(self.predictors):
            if isinstance(p, str):
                self.predictors[i] = db.load('model', p)  # type: ignore[call-overload]

    def _predict(self, X: t.Any, one: bool = False, **predict_kwargs):
        out = X
        for p in self.predictors:
            assert isinstance(p, _Predictor)
            out = p._predict(out, one=one, **predict_kwargs)
        return out



import threading
import typing as t
from contextlib import contextmanager

import click
from sqlalchemy import (
    Column,
    MetaData,
    Table,
    delete,
    insert,
    select,
)
from sqlalchemy.exc import ProgrammingError
from sqlalchemy.orm import sessionmaker

from aladindb import logging
from aladindb.backends.base.metadata import MetaDataStore, NonExistentMetadataError
from aladindb.backends.sqlalchemy.db_helper import get_db_config
from aladindb.base.serializable import Serializable
from aladindb.components.component import Component as _Component
from aladindb.misc.colors import Colors

if t.TYPE_CHECKING:
    from aladindb.backends.base.query import Select


class SQLAlchemyMetadata(MetaDataStore):
    """
    Abstraction for storing meta-data separately from primary data.

    :param conn: connection to the meta-data store
    :param name: Name to identify DB using the connection
    """

    def __init__(
        self,
        conn: t.Any,
        name: t.Optional[str] = None,
    ):
        self.name = name
        self.conn = conn
        self.dialect = conn.dialect.name
        self._init_tables()

        self._lock = threading.Lock()

    def _init_tables(self):
        # Get the DB config for the given dialect
        DBConfig = get_db_config(self.dialect)

        type_string = DBConfig.type_string
        type_json_as_string = DBConfig.type_json_as_string
        type_json_as_text = DBConfig.type_json_as_text
        type_integer = DBConfig.type_integer
        type_datetime = DBConfig.type_datetime
        type_boolean = DBConfig.type_boolean

        query_id_table_args = DBConfig.query_id_table_args
        job_table_args = DBConfig.job_table_args
        parent_child_association_table_args = (
            DBConfig.parent_child_association_table_args
        )
        component_table_args = DBConfig.component_table_args
        meta_table_args = DBConfig.meta_table_args

        metadata = MetaData()
        self.query_id_table = Table(
            'query_id_table',
            metadata,
            Column('query_id', type_string, primary_key=True),
            Column('query', type_json_as_text),
            Column('model', type_string),
            *query_id_table_args,
        )

        self.job_table = Table(
            'job',
            metadata,
            Column('identifier', type_string, primary_key=True),
            Column('component_identifier', type_string),
            Column('type_id', type_string),
            Column('info', type_json_as_string),
            Column('time', type_datetime),
            Column('status', type_string),
            Column('args', type_json_as_string),
            Column('kwargs', type_json_as_text),
            Column('method_name', type_string),
            Column('stdout', type_json_as_string),
            Column('stderr', type_json_as_string),
            Column('cls', type_string),
            *job_table_args,
        )

        self.parent_child_association_table = Table(
            'parent_child_association',
            metadata,
            Column('parent_id', type_string, primary_key=True),
            Column('child_id', type_string, primary_key=True),
            *parent_child_association_table_args,
        )

        self.component_table = Table(
            'component',
            metadata,
            Column('id', type_string, primary_key=True),
            Column('identifier', type_string),
            Column('version', type_integer),
            Column('hidden', type_boolean),
            Column('type_id', type_string),
            Column('cls', type_string),
            Column('module', type_string),
            Column('dict', type_json_as_text),
            *component_table_args,
        )

        self.meta_table = Table(
            'meta',
            metadata,
            Column('key', type_string, primary_key=True),
            Column('value', type_string),
            *meta_table_args,
        )
        metadata.create_all(self.conn)

    def url(self):
        return self.conn.url + self.name

    def drop(self, force: bool = False):
        """
        Drop the metadata store.
        """
        if not force:
            if not click.confirm(
                f'{Colors.RED}[!!!WARNING USE WITH CAUTION AS YOU '
                f'WILL LOSE ALL DATA!!!]{Colors.RESET} '
                'Are you sure you want to drop all meta-data? ',
                default=False,
            ):
                logging.warn('Aborting...')
        self.query_id_table.drop(self.conn)
        self.job_table.drop(self.conn)
        self.parent_child_association_table.drop(self.conn)
        self.component_table.drop(self.conn)
        self.meta_table.drop(self.conn)

    @contextmanager
    def session_context(self):
        sm = sessionmaker(bind=self.conn)
        session = sm()
        try:
            yield session
            session.commit()
        except Exception:
            session.rollback()
            raise
        finally:
            session.close()

    # --------------- COMPONENTS -----------------

    def component_version_has_parents(
        self, type_id: str, identifier: str, version: int
    ):
        unique_id = _Component.make_unique_id(type_id, identifier, version)
        with self.session_context() as session:
            stmt = (
                select(self.parent_child_association_table)
                .where(
                    self.parent_child_association_table.c.child_id == unique_id,
                )
                .limit(1)
            )
            res = self.query_results(self.parent_child_association_table, stmt, session)
            return len(res) > 0

    def create_component(self, info: t.Dict):
        if 'hidden' not in info:
            info['hidden'] = False
        info['id'] = f'{info["type_id"]}/{info["identifier"]}/{info["version"]}'
        with self.session_context() as session:
            stmt = insert(self.component_table).values(**info)
            session.execute(stmt)

    def create_parent_child(self, parent_id: str, child_id: str):
        with self.session_context() as session:
            stmt = insert(self.parent_child_association_table).values(
                parent_id=parent_id, child_id=child_id
            )
            session.execute(stmt)

    def delete_component_version(self, type_id: str, identifier: str, version: int):
        with self.session_context() as session:
            stmt = (
                self.component_table.select()
                .where(
                    self.component_table.c.type_id == type_id,
                    self.component_table.c.identifier == identifier,
                    self.component_table.c.version == version,
                )
                .limit(1)
            )
            res = self.query_results(self.component_table, stmt, session)
            cv = res[0] if res else None
            if cv:
                stmt_delete = delete(self.component_table).where(
                    self.component_table.c.id == cv['id']
                )
                session.execute(stmt_delete)

    def _get_component(
        self,
        type_id: str,
        identifier: str,
        version: int,
        allow_hidden: bool = False,
    ):
        with self.session_context() as session:
            stmt = select(self.component_table).where(
                self.component_table.c.type_id == type_id,
                self.component_table.c.identifier == identifier,
                self.component_table.c.version == version,
            )
            if not allow_hidden:
                stmt = stmt.where(self.component_table.c.hidden == allow_hidden)

            res = self.query_results(self.component_table, stmt, session)
            return res[0] if res else None

    def get_component_version_parents(self, unique_id: str):
        with self.session_context() as session:
            stmt = select(self.parent_child_association_table).where(
                self.parent_child_association_table.c.child_id == unique_id,
            )
            res = self.query_results(self.parent_child_association_table, stmt, session)
            parents = [r['parent_id'] for r in res]
            return parents

    def get_latest_version(
        self, type_id: str, identifier: str, allow_hidden: bool = False
    ):
        with self.session_context() as session:
            stmt = (
                select(self.component_table)
                .where(
                    self.component_table.c.type_id == type_id,
                    self.component_table.c.identifier == identifier,
                    self.component_table.c.hidden == allow_hidden,
                )
                .order_by(self.component_table.c.version.desc())
                .limit(1)
            )
            res = session.execute(stmt)
            res = self.query_results(self.component_table, stmt, session)
            versions = [r['version'] for r in res]
            if len(versions) == 0:
                raise FileNotFoundError(
                    f'Can\'t find {type_id}: {identifier} in metadata'
                )
            return versions[0]

    def hide_component_version(self, type_id: str, identifier: str, version: int):
        with self.session_context() as session:
            stmt = (
                self.component_table.update()
                .where(
                    self.component_table.c.type_id == type_id,
                    self.component_table.c.identifier == identifier,
                    self.component_table.c.version == version,
                )
                .values(hidden=True)
            )
            session.execute(stmt)

    def _replace_object(self, info, identifier, type_id, version):
        with self.session_context() as session:
            stmt = (
                self.component_table.update()
                .where(
                    self.component_table.c.type_id == type_id,
                    self.component_table.c.identifier == identifier,
                    self.component_table.c.version == version,
                )
                .values(**info)
            )
            session.execute(stmt)

    def replace_component(
        self,
        info: t.Dict[str, t.Any],
        identifier: str,
        type_id: str,
        version: t.Optional[int] = None,
    ) -> None:
        if version is not None:
            version = self.get_latest_version(type_id, identifier)
        return self._replace_object(
            info=info,
            identifier=identifier,
            type_id=type_id,
            version=version,
        )

    def show_components(self, type_id: t.Optional[str] = None, **kwargs):
        with self.session_context() as session:
            stmt = select(self.component_table)
            if type_id is not None:
                stmt = stmt.where(self.component_table.c.type_id == type_id)
            res = self.query_results(self.component_table, stmt, session)
        identifiers = [data['identifier'] for data in res]
        identifiers = sorted(set(identifiers), key=lambda x: identifiers.index(x))
        return identifiers

    def show_component_versions(self, type_id: str, identifier: str):
        with self.session_context() as session:
            stmt = select(self.component_table).where(
                self.component_table.c.type_id == type_id,
                self.component_table.c.identifier == identifier,
            )
            res = self.query_results(self.component_table, stmt, session)
            versions = [data['version'] for data in res]
            versions = sorted(set(versions), key=lambda x: versions.index(x))
            return versions

    def _update_object(
        self,
        identifier: str,
        type_id: str,
        key: str,
        value: t.Any,
        version: int,
    ):
        with self.session_context() as session:
            stmt = (
                self.component_table.update()
                .where(
                    self.component_table.c.type_id == type_id,
                    self.component_table.c.identifier == identifier,
                    self.component_table.c.version == version,
                )
                .values({key: value})
            )
            session.execute(stmt)

    # --------------- JOBS -----------------

    def create_job(self, info: t.Dict):
        with self.session_context() as session:
            stmt = insert(self.job_table).values(**info)
            session.execute(stmt)

    def get_job(self, job_id: str):
        with self.session_context() as session:
            stmt = (
                select(self.job_table)
                .where(self.job_table.c.identifier == job_id)
                .limit(1)
            )
            res = self.query_results(self.job_table, stmt, session)
            return res[0] if res else None

    def listen_job(self, identifier: str):
        # Not supported currently
        raise NotImplementedError

    def show_jobs(self):
        with self.session_context() as session:
            stmt = select(self.job_table)
            res = self.query_results(self.job_table, stmt, session)
            return [r['identifier'] for r in res]

    def update_job(self, job_id: str, key: str, value: t.Any):
        with self.session_context() as session:
            stmt = (
                self.job_table.update()
                .where(self.job_table.c.identifier == job_id)
                .values({key: value})
            )
            session.execute(stmt)

    def write_output_to_job(self, identifier, msg, stream):
        # Not supported currently
        raise NotImplementedError

    # --------------- METADATA -----------------

    def create_metadata(self, key, value):
        with self.session_context() as session:
            stmt = insert(self.meta_table).values(key=key, value=value)
            session.execute(stmt)

    def get_metadata(self, key):
        with self.session_context() as session:
            stmt = select(self.meta_table).where(self.meta_table.c.key == key).limit(1)
            res = self.query_results(self.meta_table, stmt, session)
            value = res[0]['value'] if res else None
            return value

    def update_metadata(self, key, value):
        with self.session_context() as session:
            stmt = (
                self.meta_table.update()
                .where(self.meta_table.c.key == key)
                .values({key: value})
            )
            session.execute(stmt)

    # --------------- Query ID -----------------
    def add_query(self, query: 'Select', model: str):
        query_hash = str(hash(query))

        with self.session_context() as session:
            with self._lock:
                row = {
                    'query': query.serialize(),
                    'model': model,
                    'query_id': query_hash,
                }

            stmt = insert(self.query_id_table).values(**row)
            session.execute(stmt)

    def get_query(self, query_hash: str):
        """
        Get the query from the query table corresponding to the query hash
        """
        try:
            with self.session_context() as session:
                stmt = (
                    select(self.query_id_table)
                    .where(self.query_id_table.c.query_id == str(query_hash))
                    .limit(1)
                )
                res = self.query_results(self.query_id_table, stmt, session)
                out = res[0] if res else None
        except AttributeError as e:
            if 'NoneType' in str(e):
                raise NonExistentMetadataError(
                    f'Query hash {query_hash} does not exist'
                )
            raise e

        if out is None:
            raise NonExistentMetadataError(f'Query hash {query_hash} does not exist')

    def get_model_queries(self, model: str):
        """
        Get queries related to the given model.
        """
        with self.session_context() as session:
            stmt = select(self.query_id_table).where(
                self.query_id_table.c.model == model
            )
            queries = self.query_results(self.query_id_table, stmt, session)

            unpacked_queries = []
            for row in queries:
                id = row['query_id']
                serialized = row['query']
                query = Serializable.deserialize(serialized)
                unpacked_queries.append(
                    {'query_id': id, 'query': query, 'sql': query.repr_()}
                )
            return unpacked_queries

    def disconnect(self):
        """
        Disconnect the client
        """

        # TODO: implement me

    def query_results(self, table, statment, session):
        # Some databases don't support defining statment outside of session
        result = session.execute(statment)
        columns = [col.name for col in table.columns]
        results = []
        try:
            for row in result:
                if len(row) != len(columns):
                    raise ValueError(
                        f'Number of columns in result ({row}) does not match '
                        f'number of columns in table ({columns})'
                    )
                results.append(dict(zip(columns, row)))
        except ProgrammingError:
            # Some databases don't support return empty results, such as duckdb
            return []

        return results



import os
import shutil
import typing as t
import uuid

import click

from aladindb import logging
from aladindb.backends.base.artifact import ArtifactStore
from aladindb.misc.colors import Colors


class FileSystemArtifactStore(ArtifactStore):
    """
    Abstraction for storing large artifacts separately from primary data.

    :param conn: root directory of the artifact store
    :param name: subdirectory to use for this artifact store
    """

    def __init__(
        self,
        conn: t.Any,
        name: t.Optional[str] = None,
    ):
        self.name = name
        self.conn = conn
        if not os.path.exists(self.conn):
            logging.info('Creating artifact store directory')
            os.makedirs(self.conn, exist_ok=True)

    def url(self):
        return self.conn

    def delete(self, file_id: str):
        """
        Delete artifact from artifact store
        :param file_id: File id uses to identify artifact in store
        """
        os.remove(f'{self.conn}/{file_id}')

    def drop(self, force: bool = False):
        """
        Drop the artifact store.
        """
        if not force:
            if not click.confirm(
                f'{Colors.RED}[!!!WARNING USE WITH CAUTION AS YOU '
                f'WILL LOSE ALL DATA!!!]{Colors.RESET} '
                'Are you sure you want to drop all artifacts? ',
                default=False,
            ):
                logging.warn('Aborting...')
        shutil.rmtree(self.conn, ignore_errors=force)

    def save_artifact(self, serialized: bytes) -> t.Any:
        h = uuid.uuid4().hex
        with open(os.path.join(self.conn, h), 'wb') as f:
            f.write(serialized)
        return h

    def load_bytes(self, file_id: str) -> bytes:
        with open(os.path.join(self.conn, file_id), 'rb') as f:
            return f.read()

    def disconnect(self):
        """
        Disconnect the client
        """

        # TODO: implement me



from __future__ import annotations

import dataclasses as dc
import io
import typing as t
from contextlib import contextmanager
from functools import cached_property

import torch
from torch.utils import data
from torch.utils.data import DataLoader
from tqdm import tqdm

import aladindb as s
from aladindb import logging
from aladindb.backends.base.query import Select
from aladindb.backends.query_dataset import QueryDataset
from aladindb.base.artifact import Artifact
from aladindb.base.datalayer import Datalayer
from aladindb.base.document import Document
from aladindb.base.serializable import Serializable
from aladindb.components.dataset import Dataset
from aladindb.components.encoder import Encodable
from aladindb.components.metric import Metric
from aladindb.components.model import Model, _TrainingConfiguration
from aladindb.ext.torch.utils import device_of, eval, to_device


class BasicDataset(data.Dataset):
    """
    Basic database iterating over a list of documents and applying a transformation

    :param documents: documents
    :param transform: function
    """

    def __init__(self, documents, transform):
        super().__init__()
        self.documents = documents
        self.transform = transform

    def __len__(self):
        return len(self.documents)

    def __getitem__(self, item):
        document = self.documents[item]
        if isinstance(document, Document):
            document = document.unpack()
        elif isinstance(document, Encodable):
            document = document.x
        return self.transform(document)


@dc.dataclass
class TorchTrainerConfiguration(_TrainingConfiguration):
    """
    Configuration for the PyTorch trainer.

    :param objective: Objective function
    :param loader_kwargs: Kwargs for the dataloader
    :param max_iterations: Maximum number of iterations
    :param no_improve_then_stop: Number of iterations to wait for improvement
                                 before stopping
    :param splitter: Splitter for the data
    :param download: Whether to download the data
    :param validation_interval: How often to validate
    :param listen: Which metric to listen to for early stopping
    :param optimizer_cls: Optimizer class
    :param optimizer_kwargs: Kwargs for the optimizer
    :param target_preprocessors: Preprocessors for the target
    """

    objective: t.Optional[t.Union[Artifact, t.Callable]] = None
    loader_kwargs: t.Dict = dc.field(default_factory=dict)
    max_iterations: int = 10**100
    no_improve_then_stop: int = 5
    splitter: t.Optional[Artifact] = None
    download: bool = False
    validation_interval: int = 100
    listen: str = 'objective'
    optimizer_cls: Artifact = Artifact(torch.optim.Adam, serializer='pickle')
    optimizer_kwargs: t.Dict = dc.field(default_factory=dict)
    target_preprocessors: t.Optional[t.Union[Artifact, t.Dict]] = None

    def __post_init__(self):
        super().__post_init__()
        if self.objective and not isinstance(self.objective, Artifact):
            self.objective = Artifact(artifact=self.objective)

        if self.target_preprocessors and not isinstance(
            self.target_preprocessors, Artifact
        ):
            self.target_preprocessors = Artifact(artifact=self.target_preprocessors)


@dc.dataclass(kw_only=True)
class TorchModel(Model):
    is_batch: t.Optional[t.Union[Artifact, t.Callable]] = None
    num_directions: int = 2
    optimizer_state: t.Optional[Artifact] = None
    forward_method: str = '__call__'
    train_forward_method: str = '__call__'

    def __post_init__(self):
        super().__post_init__()

        if self.model_to_device_method:
            s.logging.debug(
                f'{self.model_to_device_method} will be overriden with `to`'
            )

        self.model_to_device_method = 'to'

        self.object.serializer = 'torch'

        if self.optimizer_state is not None:
            self.optimizer.load_state_dict(self.optimizer_state.artifact)

        self._validation_set_cache = {}

    def to(self, device):
        self.object.artifact.to(device)

    @cached_property
    def optimizer(self):
        return self.training_configuration.optimizer_cls.artifact(
            self.object.artifact.parameters(),
            **self.training_configuration.optimizer_kwargs,
        )

    @property
    def optimizers(self) -> t.List:
        return [self.optimizer]

    def save(self, database: Datalayer):
        self.optimizer_state = Artifact(self.optimizer.state_dict(), serializer='torch')
        database.replace(object=self, upsert=True)

    @contextmanager
    def evaluating(self):
        yield eval(self)

    def train(self):
        return self.object.artifact.train()

    def eval(self):
        return self.object.artifact.eval()

    def parameters(self):
        return self.object.parameters()

    def state_dict(self):
        return self.object.state_dict()

    @contextmanager
    def saving(self):
        with super().saving():
            was_training = self.object.training
            try:
                self.object.eval()
                yield
            finally:
                if was_training:
                    self.object.train()

    def __getstate__(self):
        state = self.__dict__.copy()
        if isinstance(self.object, torch.jit.ScriptModule) or isinstance(
            self.object, torch.jit.ScriptFunction
        ):
            f = io.BytesIO()
            torch.jit.save(self.object, f)
            state['_object_bytes'] = f.getvalue()
        return state

    def __setstate__(self, state):
        keys = state.keys()
        for k in keys:
            if k != '_object_bytes':
                self.__dict__[k] = state[k]
            else:
                state.__dict__['object'] = torch.jit.load(
                    io.BytesIO(state.pop('object_bytes'))
                )

    def _predict_one(self, x):
        with torch.no_grad(), eval(self.object.artifact):
            if self.preprocess is not None:
                x = self.preprocess.artifact(x)
            x = to_device(x, device_of(self.object.artifact))
            singleton_batch = create_batch(x)
            method = getattr(self.object.artifact, self.forward_method)
            output = method(singleton_batch)
            output = to_device(output, 'cpu')
            args = unpack_batch(output)[0]
            if self.postprocess is not None:
                args = self.postprocess.artifact(args)
            return args

    def _predict(self, x, one: bool = False, **kwargs):  # type: ignore[override]
        assert isinstance(self.object, Artifact)
        with torch.no_grad(), eval(self.object.artifact):
            if one:
                return self._predict_one(x)
            if not isinstance(self.preprocess, Artifact):

                def func(x):
                    return x

            else:
                func = self.preprocess.artifact
            inputs = BasicDataset(x, func)
            loader = torch.utils.data.DataLoader(inputs, **kwargs)
            out = []
            for batch in tqdm(loader, total=len(loader)):
                batch = to_device(batch, device_of(self.object.artifact))

                method = getattr(self.object.artifact, self.forward_method)
                tmp = method(batch)
                tmp = to_device(tmp, 'cpu')
                tmp = unpack_batch(tmp)
                if isinstance(self.postprocess, Artifact):
                    tmp = [self.postprocess.artifact(t) for t in tmp]
                out.extend(tmp)
            return out

    def train_forward(self, X, y=None):
        X = X.to(self.device)
        if y is not None:
            y = y.to(self.device)

        method = getattr(self.object.artifact, self.train_forward_method)
        if hasattr(self.object.artifact, 'train_forward'):
            if y is None:
                return method(X)
            else:
                return method(X, y=y)
        else:
            if y is None:
                return (method(X),)
            else:
                return [method(X), y]

    def stopping_criterion(self, iteration):
        max_iterations = self.training_configuration.max_iterations
        no_improve_then_stop = self.training_configuration.no_improve_then_stop
        if isinstance(max_iterations, int) and iteration >= max_iterations:
            return True
        if isinstance(no_improve_then_stop, int):
            if self.training_configuration.listen == 'objective':
                to_listen = [-x for x in self.metric_values['objective']]
            else:
                to_listen = self.metric_values[self.training_configuration.listen]

            if max(to_listen[-no_improve_then_stop:]) < max(to_listen):
                logging.info('early stopping triggered!')
                return True
        return False

    def saving_criterion(self):
        if self.training_configuration.listen == 'objective':
            to_listen = [-x for x in self.metric_values['objective']]
        else:
            to_listen = self.metric_values[self.training_configuration.listen]
        if all([to_listen[-1] >= x for x in to_listen[:-1]]):
            return True
        return False

    def _fit(
        self,
        X: t.Any,
        *,
        y: t.Optional[t.Any] = None,
        configuration: TorchTrainerConfiguration,  # type: ignore[override]
        data_prefetch: bool = False,
        db: t.Optional[Datalayer] = None,
        metrics: t.Optional[t.Sequence[Metric]] = None,
        select: t.Optional[t.Union[Select, t.Dict]] = None,
        validation_sets: t.Optional[t.Sequence[t.Union[str, Dataset]]] = None,
    ):
        if configuration is not None:
            self.training_configuration = configuration
        if isinstance(select, dict):
            self.training_select = Serializable.deserialize(select)
        else:
            self.training_select = select
        if validation_sets is not None:
            self.validation_sets = validation_sets
        if metrics is not None:
            self.metrics = metrics

        self.train_X = X
        self.train_y = y

        train_data, valid_data = self._get_data(db=db)

        loader_kwargs = self.training_configuration.loader_kwargs
        train_dataloader = DataLoader(train_data, **loader_kwargs)
        valid_dataloader = DataLoader(valid_data, **loader_kwargs)

        if db is None:
            raise ValueError('db cannot be None')
        return self._fit_with_dataloaders(
            train_dataloader,
            valid_dataloader,
            db=db,
            validation_sets=validation_sets or [],
        )

    def log(self, **kwargs):
        out = ''
        for k, v in kwargs.items():
            if isinstance(v, dict):
                for kk, vv in v.items():
                    out += f'{k}/{kk}: {vv}; '
            else:
                out += f'{k}: {v}; '
        logging.info(out)

    def forward(self, X):
        return self.object.artifact(X)

    def extract_batch_key(self, batch, key: t.Union[t.List[str], str]):
        if isinstance(key, str):
            return batch[key]
        return [batch[k] for k in key]

    def extract_batch(self, batch):
        if self.train_y is not None:
            return [
                self.extract_batch_key(batch, self.train_X),
                self.extract_batch_key(batch, self.train_y),
            ]
        return [self.extract_batch_key(batch, self.train_X)]

    def take_step(self, batch, optimizers):
        batch = self.extract_batch(batch)
        outputs = self.train_forward(*batch)
        objective_value = self.training_configuration.objective.artifact(*outputs)
        for opt in optimizers:
            opt.zero_grad()
        objective_value.backward()
        for opt in optimizers:
            opt.step()
        return objective_value

    def compute_validation_objective(self, valid_dataloader):
        objective_values = []
        with self.evaluating(), torch.no_grad():
            for batch in valid_dataloader:
                batch = self.extract_batch(batch)
                objective_values.append(
                    self.training_configuration.objective.artifact(
                        *self.train_forward(*batch)
                    ).item()
                )
            return sum(objective_values) / len(objective_values)

    def _fit_with_dataloaders(
        self,
        train_dataloader: DataLoader,
        valid_dataloader: DataLoader,
        db: Datalayer,
        validation_sets: t.Optional[t.Sequence[t.Union[str, Dataset]]] = None,
    ):
        if validation_sets is None:
            validation_sets = []
        self.train()
        iteration = 0

        assert isinstance(self.training_configuration, TorchTrainerConfiguration)

        while True:
            for batch in train_dataloader:
                train_objective = self.take_step(batch, self.optimizers)
                self.log(fold='TRAIN', iteration=iteration, objective=train_objective)

                if iteration % self.training_configuration.validation_interval == 0:
                    valid_loss = self.compute_validation_objective(valid_dataloader)
                    all_metrics = {}
                    assert isinstance(self.metrics, list)
                    assert all(isinstance(m, Metric) for m in self.metrics)
                    metrics = [t.cast(Metric, m) for m in self.metrics]
                    for vs in validation_sets:
                        m = self._validate(
                            db=db,
                            validation_set=vs,
                            metrics=metrics,
                        )
                        all_metrics.update(m)
                    all_metrics.update({'objective': valid_loss})

                    self.append_metrics(all_metrics)
                    self.log(fold='VALID', iteration=iteration, **all_metrics)
                    if self.saving_criterion():
                        self.save(db)
                    stop = self.stopping_criterion(iteration)
                    if stop:
                        return
                iteration += 1

    def train_preprocess(self):
        preprocessors = {}
        if isinstance(self.train_X, str):
            preprocessors[self.train_X] = (
                self.preprocess if self.preprocess else lambda x: x
            )
        else:
            for model, X in zip(self.models, self.train_X):
                preprocessors[X] = (
                    model.preprocess if model.preprocess is not None else lambda x: x
                )
        if self.train_y is not None:
            if (
                isinstance(self.train_y, str)
                and self.training_configuration.target_preprocessors
            ):
                preprocessors[
                    self.train_y
                ] = self.training_configuration.target_preprocessors.get(
                    self.train_y, lambda x: x
                )
            elif isinstance(self.train_y, str):
                preprocessors[self.train_y] = lambda x: x
            elif (
                isinstance(self.train_y, list)
                and self.training_configuration.target_preprocessors
            ):
                for y in self.train_y:
                    preprocessors[
                        y
                    ] = self.training_configuration.target_preprocessors.get(
                        y, lambda x: x
                    )

        for k in preprocessors:
            if isinstance(preprocessors[k], Artifact):
                preprocessors[k] = preprocessors[k].artifact
        return lambda r: {k: preprocessors[k](r[k]) for k in preprocessors}

    def _get_data(self, db: t.Optional[Datalayer]):
        if self.training_select is None:
            raise ValueError('self.training_select cannot be None')
        train_data = QueryDataset(
            select=self.training_select,
            keys=self.training_keys,
            fold='train',
            transform=self.train_preprocess(),
            db=db,
        )
        valid_data = QueryDataset(
            select=self.training_select,
            keys=self.training_keys,
            fold='valid',
            transform=self.train_preprocess(),
            db=db,
        )
        return train_data, valid_data


def unpack_batch(args):
    """
    Unpack a batch into lines of tensor output.

    :param args: a batch of model outputs

    >>> unpack_batch(torch.randn(1, 10))[0].shape
    torch.Size([10])
    >>> out = unpack_batch([torch.randn(2, 10), torch.randn(2, 3, 5)])
    >>> type(out)
    <class 'list'>
    >>> len(out)
    2
    >>> out = unpack_batch({'a': torch.randn(2, 10), 'b': torch.randn(2, 3, 5)})
    >>> [type(x) for x in out]
    [<class 'dict'>, <class 'dict'>]
    >>> out[0]['a'].shape
    torch.Size([10])
    >>> out[0]['b'].shape
    torch.Size([3, 5])
    >>> out = unpack_batch({'a': {'b': torch.randn(2, 10)}})
    >>> out[0]['a']['b'].shape
    torch.Size([10])
    >>> out[1]['a']['b'].shape
    torch.Size([10])
    """

    if isinstance(args, torch.Tensor):
        return [args[i] for i in range(args.shape[0])]

    if isinstance(args, list) or isinstance(args, tuple):
        tmp = [unpack_batch(x) for x in args]
        batch_size = len(tmp[0])
        return [[x[i] for x in tmp] for i in range(batch_size)]

    if isinstance(args, dict):
        tmp = {k: unpack_batch(v) for k, v in args.items()}
        batch_size = len(next(iter(tmp.values())))
        return [{k: v[i] for k, v in tmp.items()} for i in range(batch_size)]

    raise NotImplementedError


def create_batch(args):
    """
    Create a singleton batch in a manner similar to the PyTorch dataloader

    :param args: single data point for batching

    >>> create_batch(3.).shape
    torch.Size([1])
    >>> x, y = create_batch([torch.randn(5), torch.randn(3, 7)])
    >>> x.shape
    torch.Size([1, 5])
    >>> y.shape
    torch.Size([1, 3, 7])
    >>> d = create_batch(({'a': torch.randn(4)}))
    >>> d['a'].shape
    torch.Size([1, 4])
    """
    if isinstance(args, (tuple, list)):
        return tuple([create_batch(x) for x in args])
    if isinstance(args, dict):
        return {k: create_batch(args[k]) for k in args}
    if isinstance(args, torch.Tensor):
        return args.unsqueeze(0)
    if isinstance(args, (float, int)):
        return torch.tensor([args])
    raise TypeError('Only tensors and tuples of tensors recursively supported...')



import typing as t
from abc import ABC, abstractmethod

from aladindb.misc.tree import tree_find, tree_rewrite

if t.TYPE_CHECKING:
    from aladindb.base.artifact import Artifact
    from aladindb.components.component import Component


def _is_artifact(t: t.Any) -> bool:
    from aladindb.base.artifact import Artifact

    return isinstance(t, Artifact)


def _has_file_id(t: t.Any) -> bool:
    return isinstance(t, dict) and 'file_id' in t


class ArtifactStoreMixin:
    @staticmethod
    def get_artifacts(tree: t.Any) -> t.Iterator['Artifact']:
        """Yield all Artifacts in a tree

        :param tree: A tree made up of dicts and lists
        """
        yield from tree_find(tree, _is_artifact)

    @staticmethod
    def _infer_artifacts(tree: t.Any) -> t.Iterator:
        """Yield all entries keyed with 'file_id' in a tree

        :param tree: A tree made up of dicts and lists
        """

        return (t['file_id'] for t in tree_find(tree, _has_file_id))

    @staticmethod
    def replace_artifacts_with_dict(tree: t.Any, info: t.Dict[str, str]) -> t.Any:
        """Replace every Artifact in a tree with an ArtifactDesc

        :param tree: A tree made up of dicts and lists
        :param info: A dictionary mapping artifacts to file_ids
        """

        def rewrite(t):
            return {'file_id': info[t.sha1], 'sha1': t.sha1, 'serializer': t.serializer}

        return tree_rewrite(tree, _is_artifact, rewrite)

    @staticmethod
    def load_artifacts(
        tree: t.Any,
        cache: t.Dict[str, 'Artifact'],
        getter: t.Callable[[str], bytes],
        serializers: t.Dict,
    ) -> t.Any:
        """Replace ArtifactDesc dicts in a tree by Artifacts with a getter

        :param tree: A tree made up of dicts and lists
        :param getter: A function that returns `bytes` given a file_id
        :param cache: A dictionary caching artifacts by file_id
        """

        def make(file_id, serializer):
            bytes_ = getter(file_id)
            return serializers[serializer].decode(bytes_)

        return ArtifactStoreMixin._load_artifacts(tree, make, cache)

    def _replace_artifacts(self, tree: t.Any, info: t.Optional[t.Dict] = None) -> t.Any:
        """Replace every Artifact in a tree with hash of the value, looked up in Info

        :param tree: A tree made up of dicts and lists
        :param info: A dictionary mapping hashes to [TBD]
        """

        def rewrite(t):
            return info[t.sha1(self.serializers)]

        return tree_rewrite(tree, _is_artifact, rewrite)

    @staticmethod
    def _load_artifacts_from_cache(
        tree: t.Any, cache: t.Dict[str, 'Artifact']
    ) -> t.Any:
        """Replace ArtifactDesc dicts in a tree by Artifacts."""

        def rewrite(t):
            file_id = t['file_id']
            return cache[file_id]

        return tree_rewrite(tree, _has_file_id, rewrite)

    @staticmethod
    def _load_artifacts(
        tree: t.Any,
        make: t.Callable,
        cache: t.Dict[str, 'Artifact'],
        artifact_store: t.Optional['ArtifactStore'] = None,
    ) -> t.Any:
        """Replace ArtifactDesc dicts in a tree by Artifacts."""
        from aladindb.base.artifact import Artifact

        def rewrite(t):
            file_id = t['file_id']
            try:
                return cache[file_id]
            except KeyError:
                pass
            artifact = make(file_id=file_id, serializer=t['serializer'])
            cache[file_id] = Artifact(
                artifact=artifact,
                serializer=t['serializer'],
                artifact_store=artifact_store,
                file_id=file_id,
            )
            return cache[file_id]

        return tree_rewrite(tree, _has_file_id, rewrite)


class ArtifactStore(ABC, ArtifactStoreMixin):
    """
    Abstraction for storing large artifacts separately from primary data.

    :param conn: connection to the meta-data store
    :param name: Name to identify DB using the connection
    """

    def __init__(
        self,
        conn: t.Any,
        name: t.Optional[str] = None,
    ):
        self.name = name
        self.conn = conn
        self._serializers = None

    @property
    def serializers(self):
        assert self._serializers is not None, 'Serializers not initialized!'
        return self._serializers

    @serializers.setter
    def serializers(self, value):
        self._serializers = value

    @abstractmethod
    def url(self):
        """
        Artifact store connection url
        """
        pass

    @abstractmethod
    def delete(self, file_id: str):
        """
        Delete artifact from artifact store
        :param file_id: File id uses to identify artifact in store
        """
        pass

    @abstractmethod
    def drop(self, force: bool = False):
        """
        Drop the artifact store.

        :param force: If ``True``, don't ask for confirmation
        """
        pass

    def create(self, bytes: t.Any):
        """
        Save serialized object in the artifact store.

        :param object: Object to serialize
        :param bytes: Serialized object
        """
        return self.save_artifact(bytes)

    @abstractmethod
    def save_artifact(self, serialized: bytes) -> t.Any:
        """
        Save serialized object in the artifact store.

        :param serialized: Serialized object
        """
        pass

    @abstractmethod
    def load_bytes(self, file_id: str) -> bytes:
        """
        Load bytes from artifact store.

        :param file_id: Identifier of artifact in the store
        """
        pass

    def load_artifact(self, file_id: str, serializer: str) -> t.Any:
        """
        Load artifact from artifact store, and deserialize.

        :param file_id: Identifier of artifact in the store
        :param serializer: Serializer to use for deserialization
        """
        bytes = self.load_bytes(file_id)
        serializer_function = self.serializers[serializer]
        return serializer_function.decode(bytes)

    def save(self, artifacts: t.Iterator['Artifact']) -> t.Dict:
        """
        Save list of artifacts and replace the artifacts with file reference
        :param artifacts: List of ``Artifact`` instances
        """

        artifact_details = dict()
        for a in artifacts:
            file_detail = a.save(self)
            artifact_details[file_detail['sha1']] = file_detail
        return artifact_details

    def replace(self, serialized: t.Dict, info: t.Optional[t.Dict] = None) -> t.Dict:
        """Replace every Artifact in the given ``object`` serialized info with
        artifact lookup.

        :param serialized: Serialized component dict
        :param artifact_info: A dictionary mapping hashes to [TBD]
        """
        return t.cast(t.Dict, self._replace_artifacts(serialized, info))

    def update(self, object: 'Component', metadata_info: t.Dict = {}):
        """
        Update artifacts with the updated object in the artifact store.
        :param object: A ``Component`` instance
        :param metadata_info: Previous object info from metadata store
        """
        serialized, artifacts = object.serialized
        artifact_details = self.save(artifacts)
        artifacts = tuple(set(self._infer_artifacts(metadata_info)))
        for oa in artifacts:
            self.delete(oa)

        return self.replace(serialized, artifact_details)

    def load(self, info, cache: t.Dict = {}, lazy: bool = False):
        """
        Recursively search through a tree for ArtifactDescs by an Artifact with
        that `file_id`.

        :param info: A tree made up of dicts and lists
        :param cache: A cache dictionary of Artifacts, keyed by file_id
        :param lazy: If True, don't load the artifacts; used so that big objects don't
                     clutter up the memory.
        """

        def no_load(*args, **kwargs):
            pass

        make = t.cast(t.Callable, self.load_artifact if not lazy else no_load)

        return self._load_artifacts(
            tree=info,
            make=make,
            cache=cache,
            artifact_store=self,
        )

    @staticmethod
    def load_from_cache(info, cache: t.Dict = {}):
        """
        Recursively search through a tree for ArtifactDescs by an Artifact with
        that `file_id` from given cache.

        :param info: A tree made up of dicts and lists
        :param cache: A cache dictionary of Artifacts, keyed by file_id
        """
        return ArtifactStoreMixin._load_artifacts_from_cache(info, cache)

    @abstractmethod
    def disconnect(self):
        """
        Disconnect the client
        """



import dataclasses as dc
import typing as t

import sentence_transformers

from aladindb.base.artifact import Artifact
from aladindb.components.model import Model


@dc.dataclass(kw_only=True)
class SentenceTransformer(Model):
    object: t.Union[Artifact, t.Callable, None] = None

    def __post_init__(self):
        super().__post_init__()
        if self.object is None:
            self.object = Artifact(
                artifact=sentence_transformers.SentenceTransformer(
                    self.identifier, device=self.device
                )
            )
        self.object.artifact = self.object.artifact.to(self.device)
        self.model_to_device_method = '_to'

    def _to(self, device):
        self.object.artifact = self.object.artifact.to(device)
        self.object.artifact._target_device = device
